# vim: ts=4:sts=4:sw=4
#
# @date 2022-02-04
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import re
import typing
import logging
import datetime

import defusedxml.ElementTree as ET
import numpy
import numpy.typing
import tifffile
from scipy.interpolate import RectBivariateSpline
from scipy.stats import scoreatpercentile
import pyproj

import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import Granule, Extrema

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'
SUPPORTED_VARIABLES = ['roughness', 'sigma0', 'incidence']
SUPPORTED_CONTRASTS = ['sea', 'ice', 'relative', 'relative_strict']


class InputPathMissing(IOError):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


class S1ProductAnnotationNotFound(IOError):
    """Error raised when the path for the S-1 product annotation file does not
    exist."""
    pass


class S1CalibrationAnnotationNotFound(IOError):
    """Error raised when the path for the S-1 calibration annotation file does
    not exist."""
    pass


class S1VariableNotSupported(ValueError):
    """Error raised when the S-1 variable is not supported."""
    pass


class S1GeolocationGridNotAsExpected(Exception):
    """Error raised when the S-1 geolocation grid is not as expected."""
    pass


class S1ProductNotKnown(ValueError):
    """Error raised when the annotated S-1 product is not known."""
    pass


class S1ModeNotKnown(ValueError):
    """Error raised when the annotated S-1 mode is not known."""
    pass


class S1ProductNotSupported(Exception):
    """Error raised when the S1 input product is not supported (IW SLC or EW
    SLC)."""
    pass


class ContrastNameNotKnown(ValueError):
    """Error raised when the contrast name is not known."""
    pass


def read_product_annotations(prod_annot_path: str
                             ) -> typing.Dict[str, typing.Any]:
    """"""
    annot = {}
    tree = ET.parse(prod_annot_path)
    ads = tree.getroot()

    # Header
    adsh = ads.find('./adsHeader')
    annot['mission'] = adsh.find('./missionId').text
    annot['product'] = adsh.find('./productType').text
    annot['polarisation'] = adsh.find('./polarisation').text
    annot['mode'] = adsh.find('./mode').text
    annot['swath'] = adsh.find('./swath').text
    annot['start_time'] = adsh.find('./startTime').text
    annot['stop_time'] = adsh.find('./stopTime').text

    # Product info
    adspi = ads.find('./generalAnnotation/productInformation')
    annot['pass'] = adspi.find('./pass').text

    # Image info
    adsii = ads.find('./imageAnnotation/imageInformation')
    range_pixel_spacing_str = adsii.find('./rangePixelSpacing').text
    annot['range_pixel_spacing'] = float(range_pixel_spacing_str)
    annot['azimuth_pixel_spacing'] = \
        float(adsii.find('./azimuthPixelSpacing').text)
    annot['incidence_angle_mid_swath'] = \
        float(adsii.find('./incidenceAngleMidSwath').text)
    annot['number_of_samples'] = int(adsii.find('./numberOfSamples').text)
    annot['number_of_lines'] = int(adsii.find('./numberOfLines').text)

    # Geolocation grid
    adsgg = ads.find('./geolocationGrid/geolocationGridPointList')
    points = adsgg.findall('./geolocationGridPoint')
    npoints = len(points)
    # azit = []
    # rant = []
    _lin = []
    _pix = []
    _lat = []
    _lon = []
    _hei = []
    _inc = []
    # elev = []
    for point in points:
        # azit.append(point.find('./azimuthTime').text)
        # rant.append(point.find('./slantRangeTime').text)
        _lin.append(point.find('./line').text)
        _pix.append(point.find('./pixel').text)
        _lat.append(point.find('./latitude').text)
        _lon.append(point.find('./longitude').text)
        _hei.append(point.find('./height').text)
        _inc.append(point.find('./incidenceAngle').text)
        # elev.append(point.find('./elevationAngle').text)
    lin = numpy.array(_lin, dtype='int32')
    pix = numpy.array(_pix, dtype='int32')
    lat = numpy.array(_lat, dtype='float32')
    lon = numpy.array(_lon, dtype='float32')
    hei = numpy.array(_hei, dtype='float32')
    inc = numpy.array(_inc, dtype='float32')
    npixels = len(numpy.where(lin == lin[0])[0])
    nlines = len(numpy.where(pix == pix[0])[0])
    if npixels * nlines != npoints:
        raise S1GeolocationGridNotAsExpected()
    shape = (nlines, npixels)
    lin = lin.reshape(shape)
    pix = pix.reshape(shape)
    chk = lin.min(axis=1) != lin.max(axis=1)
    if chk.any():
        raise S1GeolocationGridNotAsExpected()
    chk = pix.min(axis=0) != pix.max(axis=0)
    if chk.any():
        raise S1GeolocationGridNotAsExpected()
    geoloc = {}
    geoloc['line'] = lin
    geoloc['pixel'] = pix
    geoloc['latitude'] = lat.reshape(shape)
    geoloc['longitude'] = lon.reshape(shape)
    geoloc['height'] = hei.reshape(shape)
    geoloc['incidence_angle'] = inc.reshape(shape)
    linargsort = geoloc['line'][:, 0].argsort()
    if ((linargsort[1:] - linargsort[0:-1]) != 1).any():
        logging.warning('Geolocation lines are not sorted -> do sort')
        for key in geoloc.keys():
            geoloc[key] = geoloc[key][linargsort, :]
    pixargsort = geoloc['pixel'][0, :].argsort()
    if ((pixargsort[1:] - pixargsort[0:-1]) != 1).any():
        logging.warning('Geolocation pixels are not sorted -> do sort')
        for key in geoloc.keys():
            geoloc[key] = geoloc[key][:, pixargsort]
    annot['geolocation_grid'] = geoloc

    return annot


def read_calibration_annotations(cal_annot_path: str
                                 ) -> typing.Dict[str, typing.Any]:
    """"""
    annot = {}
    tree = ET.parse(cal_annot_path)
    ads = tree.getroot()

    # Look-up tables for calibration
    vectors = ads.findall('./calibrationVectorList/calibrationVector')
    # azit = []
    lin = []
    pix = []
    sig = []
    # bet = []
    # gam = []
    # dn = []
    for vector in vectors:
        # azit.append(vector.find('./azimuthTime').text)
        lin.append(vector.find('./line').text)
        pix.append(vector.find('./pixel').text.split())
        sig.append(vector.find('./sigmaNought').text.split())
        # bet.append(vector.find('./betaNought').text.split())
        # gam.append(vector.find('./gamma').text.split())
        # dn.append(vector.find('./dn').text.split())
    luts = {}
    luts['pixel'] = numpy.array(pix, dtype='int32')
    nline, npixel = luts['pixel'].shape
    lin_array = numpy.array(lin, dtype='int32')
    luts['line'] = numpy.tile(lin_array[:, numpy.newaxis], (1, npixel))
    luts['sigma0'] = numpy.array(sig, dtype='float32')
    annot['luts'] = luts

    return annot


def read_tiff(tiff_path: str) -> numpy.typing.NDArray:
    """"""
    data = tifffile.imread(tiff_path, mode='rb')
    return data


def read_tiff_as_memmap(tiff_path: str) -> numpy.typing.NDArray:
    """"""
    data = tifffile.memmap(tiff_path, mode='r')
    return data


def interpolate_grid(xval: numpy.typing.NDArray,
                     yval: numpy.typing.NDArray,
                     zval: numpy.typing.NDArray,
                     xint: numpy.typing.NDArray,
                     yint: numpy.typing.NDArray,
                     degrees: typing.Tuple[int, int] = (3, 3),
                     blocksize: int = 25000000
                     ) -> numpy.typing.NDArray:
    """"""
    # Set bbox in case of extrapolation is needed
    bbox = []
    for xyv, xyi in zip([xval, yval], [xint, yint]):
        if xyi.min() < xyv.min():
            bbox.append(xyi.min())
        else:
            bbox.append(None)
        if xyi.max() > xyv.max():
            bbox.append(xyi.max())
        else:
            bbox.append(None)
    # Interpolate with RectBivariateSpline
    # Warning: RectBivariateSpline is called with grid=True meaning that
    # xint and yint must be sorted to increasing order.
    func = RectBivariateSpline(xval, yval, zval, bbox=bbox,
                               kx=degrees[0], ky=degrees[1])
    if blocksize is None:
        values = func(xint, yint, grid=True).astype(zval.dtype)
    else:
        ny = yint.size
        nx = xint.size
        block_ny = ny
        block_nx = numpy.maximum(blocksize // block_ny, 1)
        nblock = int(numpy.ceil(nx / float(block_nx)))
        values = numpy.empty((nx, ny), dtype=zval.dtype)
        for iblock in range(nblock):
            x0 = iblock * block_nx
            x1 = numpy.minimum((iblock + 1) * block_nx, nx)
            xsli = slice(x0, x1)
            values[xsli, :] = func(xint[xsli], yint, grid=True)
    return values


def move_geoloc_height(lon: numpy.typing.NDArray,
                       lat: numpy.typing.NDArray,
                       hei: numpy.typing.NDArray,
                       inc: numpy.typing.NDArray,
                       target_height: float = 0.
                       ) -> typing.Tuple[numpy.typing.NDArray,
                                         numpy.typing.NDArray,
                                         numpy.typing.NDArray]:
    """"""
    geod = pyproj.Geod(ellps='WGS84')
    forw, back, _ = geod.inv(lon[:, :-1], lat[:, :-1], lon[:, 1:], lat[:, 1:])
    forw = numpy.hstack((forw, forw[:, [-1]]))
    back = numpy.hstack((back[:, [0]], back))
    mvdist = (target_height - hei) / numpy.tan(numpy.deg2rad(inc))
    mvforw = forw
    indneg = numpy.where(mvdist < 0)
    mvdist[indneg] = -mvdist[indneg]
    mvforw[indneg] = back[indneg]
    _lon, _lat, _ = geod.fwd(lon, lat, mvforw, mvdist)
    _hei = numpy.full_like(hei, target_height)
    return _lon, _lat, _hei


def s1_extrema(var_dict: typing.Dict[str, typing.Any],
               default_min: float,
               default_max: float) -> Extrema:
    """"""
    shp = var_dict['array'].shape
    noborder = (slice(int(shp[0] * 0.05), int(shp[0] * 0.95)),
                slice(int(shp[1] * 0.1), int(shp[1] * 0.9)))
    _min = scoreatpercentile(var_dict['array'][noborder], 0.1)
    _max = scoreatpercentile(var_dict['array'][noborder], 99.9)
    return _min, _max


def s1_extrema_strict(var_dict: typing.Dict[str, typing.Any],
                      default_min: float,
                      default_max: float) -> Extrema:
    """"""
    _min = var_dict['array'].min()
    _max = var_dict['array'].max()
    return _min, _max


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    supp_vars = ', '.join(SUPPORTED_VARIABLES)
    supp_ctst = ', '.join(SUPPORTED_CONTRASTS)
    inp = ('    path\tPath of the input file',
           '    variables\tComma-separated list of variable identifiers to'
           ' extract from the input file. Identifiers of supported variables'
           ' are {}.'.format(supp_vars),
           '    variable_overrides_VAR\tDefinition of variable attributes'
           ' that are either missing or erroneous in the input file.'
           ' VAR is the identifier of the variable and the value of this'
           ' option is a comma-separated list of key:value. For instance'
           ' here, variable_overrides_roughness=valid_max:3 allows to'
           ' directly set valid_max of roughness variable.',
           '    contrast\tStrategy to automatically set valid_min/valid_max'
           ' of roughness variable. Supported contrasts are'
           ' {}.'.format(supp_ctst),
           '    spacing\tComma-separated list (length 2) of extraction spacing'
           ' in pixel. For instance, spacing=4,3 gives a downsampled version'
           ' by 4 in azimuth and by 3 in range.',
           '    mspacing\tComma-separated list (length 2) of extraction'
           ' spacing in meter. For instance, mspacing=100,50 gives a'
           ' downsampled version with 100m spacing in azimuth and 50m spacing'
           ' in range. Ignored if spacing is given.',
           )
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def set_extrema(granule: Granule,
                vars_id: typing.Sequence[str],
                annot: typing.Dict[str, typing.Any],
                contrast: typing.Optional[str],
                float_variables: typing.Sequence[str]
                ) -> typing.Optional[str]:
    """"""
    # Sigma0
    if 'sigma0' in vars_id:
        vmin = granule.vars['sigma0']['valid_min']
        vmax = granule.vars['sigma0']['valid_max']
        if vmin is None or vmax is None:
            if 'sigma0' not in float_variables:
                _vmin, _vmax = s1_extrema(granule.vars['sigma0'], 0., 0.)
            else:
                _vmin, _vmax = s1_extrema_strict(granule.vars['sigma0'],
                                                 0., 0.)
            if vmin is None:
                granule.vars['sigma0']['valid_min'] = _vmin
            if vmax is None:
                granule.vars['sigma0']['valid_max'] = _vmax
    # Incidence
    if 'incidence' in vars_id:
        vmin = granule.vars['incidence']['valid_min']
        vmax = granule.vars['incidence']['valid_max']
        if vmin is None:
            _vmin = granule.vars['incidence']['array'].min()
            granule.vars['incidence']['valid_min'] = _vmin
        if vmax is None:
            _vmax = granule.vars['incidence']['array'].max()
            granule.vars['incidence']['valid_max'] = _vmax
    # Roughness
    if 'roughness' in vars_id:
        vmin = granule.vars['roughness']['valid_min']
        vmax = granule.vars['roughness']['valid_max']
        if vmin is not None and vmax is not None:  # user choice
            # Nothing to do but we don't want contrast='relative' or
            # contrast='relative_strict'.
            # (otherwise it will trigger a transform)
            contrast = None
        else:  # default vmin and/or vmax based on contrast string
            if contrast is None:  # default contrast string
                if 'sea_surface_roughness' in float_variables:
                    contrast = 'relative_strict'
                elif annot['mode'] == 'WV':
                    contrast = 'relative'
                else:
                    contrast = 'sea'
            else:
                if contrast not in SUPPORTED_CONTRASTS:
                    raise ContrastNameNotKnown()
            if contrast == 'relative':
                # a transform will be added for vmin/vmax computation
                pass
            elif contrast == 'relative_strict':
                # a transform will be added for vmin/vmax computation
                pass
            elif contrast == 'sea':
                if annot['polarisation'] == 'VV':
                    if vmin is None:
                        vmin = 0.
                    if vmax is None:
                        vmax = 2.
                elif annot['polarisation'] == 'HH':
                    if vmin is None:
                        vmin = 0.
                    if vmax is None:
                        vmax = 2.5
                else:
                    if vmin is None:
                        vmin = 1.
                    if vmax is None:
                        vmax = 3.
            elif contrast == 'ice':
                if annot['polarisation'] in ['HH', 'VV']:
                    if vmin is None:
                        vmin = 0.
                    if vmax is None:
                        vmax = 3.5
                else:
                    if vmin is None:
                        vmin = 1.
                    if vmax is None:
                        vmax = 5.
            else:
                raise ContrastNameNotKnown()
            granule.vars['roughness']['valid_min'] = vmin
            granule.vars['roughness']['valid_max'] = vmax
    return contrast


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    # Get inputs
    _tiff_path = input_opts.get('path', None)
    if _tiff_path is None:
        raise InputPathMissing()
    tiff_path = os.path.normpath(_tiff_path)

    _vars_id = input_opts.get('variables', None)
    if _vars_id is None:
        vars_id = ['roughness']  # Default
    else:
        _vars_id = _vars_id.split(',')
        for var in _vars_id:
            if var not in SUPPORTED_VARIABLES:
                raise S1VariableNotSupported()
        vars_id = []
        for var in SUPPORTED_VARIABLES:
            if var in _vars_id:
                vars_id.append(var)

    contrast = input_opts.get('contrast', None)
    spacing = input_opts.get('spacing', None)
    if spacing is not None:
        spacing = numpy.array(spacing.split(','), dtype='int32')
    mspacing = input_opts.get('mspacing', None)
    if mspacing is not None:
        mspacing = numpy.array(mspacing.split(','), dtype='float32')

    # Init granule
    variables: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
    if 'roughness' in vars_id:
        variables['roughness'] = {'name': 'sea_surface_roughness',
                                  'long_name': 'sea surface roughness',
                                  # 'units': '', # dimensionless
                                  'valid_min': None,
                                  'valid_max': None,
                                  'options': {}}
    if 'sigma0' in vars_id or 'roughness' in vars_id:
        variables['sigma0'] = {'name': 'sigma0',
                               'long_name': 'normalized radar cross-section',
                               # 'units': '', # dimensionless
                               'valid_min': None,
                               'valid_max': None,
                               'options': {}}
    if 'incidence' in vars_id or 'roughness' in vars_id:
        variables['incidence'] = {'name': 'incidence',
                                  'long_name': 'incidence angle',
                                  'units': 'degrees',
                                  'valid_min': None,
                                  'valid_max': None,
                                  'options': {}}
    variables['lat'] = {'units': 'degrees_north',
                        'options': {}}
    variables['lon'] = {'units': 'degrees_east',
                        'options': {}}

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables
    idf_converter.lib.apply_var_overrides(input_opts, granule)

    # Check existance of required files
    tiff_dir, tiff_bname = os.path.split(tiff_path)
    annot_dir = os.path.join(os.path.dirname(tiff_dir), 'annotation')
    xml_bname = tiff_bname.replace('.tiff', '.xml')

    prod_annot_path = os.path.join(annot_dir, xml_bname)
    if not os.path.exists(prod_annot_path):
        raise S1ProductAnnotationNotFound()

    cal_annot_path = os.path.join(annot_dir, 'calibration',
                                  'calibration-{}'.format(xml_bname))
    if not os.path.exists(cal_annot_path):
        raise S1CalibrationAnnotationNotFound()

    # Read relevant annotations
    annot = read_product_annotations(prod_annot_path)
    annot.update(read_calibration_annotations(cal_annot_path))

    # Set downsampling factor
    # - grd_spacing is product ground spacing in meter
    # - spacing is target spacing in pixel (ie downsampling factor)
    # - mspacing is target spacing in meter
    azi_grd_spacing = annot['azimuth_pixel_spacing']
    ran_grd_spacing = annot['range_pixel_spacing']
    if annot['product'] == 'SLC':
        incmid = numpy.deg2rad(annot['incidence_angle_mid_swath'])
        ran_grd_spacing = ran_grd_spacing / numpy.sin(incmid)
    grd_spacing = numpy.array([azi_grd_spacing, ran_grd_spacing])
    if spacing is not None:  # user choice
        pass
    elif mspacing is not None:  # user choice
        spacing = (mspacing / grd_spacing).round().astype('int32')
    else:  # default
        if annot['product'] == 'GRD':
            spacing = numpy.array([2, 2])
        elif annot['product'] == 'SLC':
            if annot['mode'] == 'WV':
                mspacing = numpy.array([15., 15.])  # meters
            elif re.match(r'^S[1-6]$', annot['mode']) is not None:
                mspacing = numpy.array([15., 15.])  # meters
            elif annot['mode'] == 'IW':
                raise S1ProductNotSupported()
            elif annot['mode'] == 'EW':
                raise S1ProductNotSupported()
            else:
                raise S1ModeNotKnown()
            spacing = (mspacing / grd_spacing).round().astype('int32')
        else:
            raise S1ProductNotKnown()
    mspacing = spacing * grd_spacing
    frshp = numpy.array([annot['number_of_lines'],
                         annot['number_of_samples']])  # full res shape
    _frshp = frshp - numpy.mod(frshp, spacing)
    dshp = (_frshp / spacing).astype('int32')  # downsampled shape
    fullres_nrow = int(_frshp[0])
    fullres_ncell = int(_frshp[1])
    nrow = int(dshp[0])
    ncell = int(dshp[1])
    granule.dims['row'] = nrow
    granule.dims['cell'] = ncell

    # Get sigma0
    if 'sigma0' in granule.vars:
        # - Read digital number stored in tiff file and downsample it
        # -- blocksize version
        # tiff memmap is called at each iteration, otherwise memory usage
        # increases up to the tiff size.
        sigma0 = numpy.empty((nrow, ncell), dtype='float32')
        blocksize = 50000000
        block_ncell = fullres_ncell
        block_nrow = numpy.maximum(blocksize // block_ncell, spacing[0])
        block_nrow -= block_nrow % spacing[0]
        nblock = int(numpy.ceil(fullres_nrow / float(block_nrow)))
        for iblock in range(nblock):
            fr_row0 = iblock * block_nrow
            fr_row1 = numpy.minimum((iblock + 1) * block_nrow, fullres_nrow)
            fr_sli = (slice(fr_row0, fr_row1), slice(0, fullres_ncell))
            row0 = int(fr_row0 / spacing[0])
            row1 = int(fr_row1 / spacing[0])
            sli = (slice(row0, row1), slice(0, ncell))
            rshp = (row1 - row0, spacing[0], ncell, spacing[1])
            dnum: numpy.typing.NDArray
            try:
                dnum = read_tiff_as_memmap(tiff_path)
            except ValueError:
                dnum = read_tiff(tiff_path)
            _sigma0 = numpy.abs(dnum[fr_sli]).astype('float32') ** 2.
            _sigma0 = _sigma0.reshape(rshp).mean(-1).mean(1)
            sigma0[sli] = _sigma0
        # -- non blocksize version
        # sigma0 = numpy.abs(read_tiff(tiff_path)).astype('float32') ** 2.
        # if spacing.max() > 1:
        #     rshp = (nrow, spacing[0], ncell, spacing[1])
        #     frsli = (slice(0, _frshp[0]), slice(0, _frshp[1]))
        #     sigma0 = sigma0[frsli].reshape(rshp).mean(-1).mean(1)
        # - Calibrate sigma0
        # Warning: xint and yint must be sorted to increasing order
        xint = numpy.arange(nrow) * spacing[0] + (spacing[0] - 1) / 2.
        yint = numpy.arange(ncell) * spacing[1] + (spacing[1] - 1) / 2.
        xval = annot['luts']['line'][:, 0]
        yval = annot['luts']['pixel'][0, :]
        zval = annot['luts']['sigma0']
        lut_sigma0 = interpolate_grid(xval, yval, zval, xint, yint,
                                      degrees=(1, 1), blocksize=25000000)
        sigma0 /= lut_sigma0 ** 2.
        del lut_sigma0
        granule.vars['sigma0']['array'] = sigma0

    # Get incidence angle
    if 'incidence' in granule.vars:
        # Warning: xint and yint must be sorted to increasing order
        xint = numpy.arange(nrow) * spacing[0] + (spacing[0] - 1) / 2.
        yint = numpy.arange(ncell) * spacing[1] + (spacing[1] - 1) / 2.
        xval = annot['geolocation_grid']['line'][:, 0]
        yval = annot['geolocation_grid']['pixel'][0, :]
        zval = annot['geolocation_grid']['incidence_angle']
        inc_angle = interpolate_grid(xval, yval, zval, xint, yint,
                                     degrees=(1, 3), blocksize=25000000)
        granule.vars['incidence']['array'] = inc_angle

    # Set valid_min / valid_max
    _float_variables = output_opts.get('use_float32_values', '')
    float_variables = [_.strip() for _ in _float_variables.split(',')
                       if 0 < len(_.strip())]
    contrast = set_extrema(granule, vars_id, annot, contrast, float_variables)

    # Geolocation
    # - Move geoloc to 0 height
    geoloc = annot['geolocation_grid']
    lon = geoloc['longitude']
    lat = geoloc['latitude']
    hei = geoloc['height']
    inc = geoloc['incidence_angle']
    lon, lat, hei = move_geoloc_height(lon, lat, hei, inc, 0.)
    if lon.min() < -135 and lon.max() > 135:
        lon[numpy.where(lon < 0)] += 360.
    granule.vars['lat']['array'] = lat
    granule.vars['lon']['array'] = lon
    # - Settings for compute_gcps() method in swath module
    ngeorow = lat.shape[0]
    ngeocell = lat.shape[1]
    gcp_distribution = {}
    gcp_distribution['geoloc_row'] = numpy.arange(ngeorow)
    gcp_distribution['geoloc_cell'] = numpy.arange(ngeocell)
    gcp_distribution['input_row'] = (geoloc['line'][:, 0] + 0.5) / spacing[0]
    gcp_distribution['input_cell'] = (geoloc['pixel'][0, :] + 0.5) / spacing[1]
    dst_lin = numpy.linspace(0, nrow, num=ngeorow - 1, endpoint=False)
    gcp_distribution['output_row'] = dst_lin.round().astype('int32')
    dst_pix = numpy.linspace(0, ncell, num=ngeocell - 1, endpoint=False)
    gcp_distribution['output_cell'] = dst_pix.round().astype('int32')
    output_opts['gcp_distribution'] = gcp_distribution
    input_opts['geoloc_at_pixel_center'] = 'false'

    # Metadata
    safe_dir = os.path.dirname(os.path.dirname(tiff_path))
    safe_name = os.path.splitext(os.path.basename(safe_dir))[0]
    pid = safe_name.split('_')[-1]
    granule_name = '{}-{}'.format(os.path.splitext(tiff_bname)[0], pid)

    time_fmt = '%Y-%m-%dT%H:%M:%S.%f'
    start_time = datetime.datetime.strptime(annot['start_time'], time_fmt)
    stop_time = datetime.datetime.strptime(annot['stop_time'], time_fmt)

    platform = annot['mission'].replace('S1', 'Sentinel-1')

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = mspacing.min()
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_time
    granule.meta['time_coverage_end'] = stop_time
    granule.meta['institution'] = 'ESA'
    granule.meta['platform'] = platform
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'C-SAR'
    granule.meta['sensor_type'] = 'sar'
    granule.meta['sensor_mode'] = annot['mode']
    granule.meta['sensor_swath'] = annot['swath']
    granule.meta['sensor_polarisation'] = annot['polarisation']
    granule.meta['sensor_pass'] = annot['pass']
    granule.meta['safe_name'] = safe_name

    # Transforms
    transforms = []
    # - SAR roughness computation
    if 'roughness' in vars_id:
        transforms.append(('compute_roughness',
                           {'targets': ('roughness',),
                            'sigma0_var_id': 'sigma0',
                            'incidence_var_id': 'incidence',
                            'polarisation': annot['polarisation']}))
    # - SAR roughness relative valid_min / valid_max
    if 'roughness' in vars_id and contrast == 'relative':
        methods = {'roughness': s1_extrema}
        min_values = {'roughness': 0.}
        max_values = {'roughness': 0.}
        transforms.append(('extrema_methods', {'targets': ('roughness',),
                                               'methods': methods,
                                               'min_values': min_values,
                                               'max_values': max_values}))
    elif 'roughness' in vars_id and contrast == 'relative_strict':
        methods = {'roughness': s1_extrema_strict}
        min_values = {'roughness': 0.}
        max_values = {'roughness': 0.}
        transforms.append(('extrema_methods', {'targets': ('roughness',),
                                               'methods': methods,
                                               'min_values': min_values,
                                               'max_values': max_values}))
    # - masking sigma0 and roughness
    if 'roughness' in vars_id or 'sigma0' in vars_id:
        mask = granule.vars['sigma0']['array'] == 0
        if mask.max() == True:  # noqa: E712
            targets = [vid for vid in vars_id
                       if vid in ['roughness', 'sigma0']]
            transforms.append(('static_common_mask', {'targets': targets,
                                                      'mask': mask}))
    # - Remove variables
    to_remove = []
    if 'sigma0' in granule.vars and 'sigma0' not in vars_id:
        to_remove.append('sigma0')
    if 'incidence' in granule.vars and 'incidence' not in vars_id:
        to_remove.append('incidence')
    if len(to_remove) > 0:
        transforms.append(('remove_vars', {'targets': to_remove}))

    output_opts['__export'] = vars_id

    yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-10
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import pyproj
import netCDF4
import logging
import idf_converter.lib
import idf_converter.readers.netcdf_grid_yx
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import ReaderResult

logger = logging.getLogger()

DATA_MODEL = idf_converter.readers.netcdf_grid_yx.DATA_MODEL


class FilenamePatternNotSupported(ValueError):
    """Error raised when the name of the input file does nt match any known
    pattern, preventing the detecting of the time coverage, platform and
    sensor."""
    def __init__(self, file_name: str) -> None:
        """"""
        self.file_name = file_name


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    in_msg, out_msg = idf_converter.readers.netcdf_grid_yx.help()
    return (in_msg, out_msg)


def get_ice_mask(granule: Granule
                 ) -> numpy.typing.NDArray:
    """"""
    threshold = 0.00001  # 0.001%
    mask: numpy.typing.NDArray
    mask = (threshold >= numpy.ma.getdata(granule.vars['fice']['array']))
    return mask


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise idf_converter.readers.netcdf_grid_yx.InputPathMissing()
    input_path = os.path.normpath(_input_path)

    if 'time_variable' not in input_opts:
        input_opts['time_variable'] = 'time'
    if 'x_variable' not in input_opts:
        input_opts['x_variable'] = 'x'
    if 'y_variable' not in input_opts:
        input_opts['y_variable'] = 'y'
    if ('depth_value' in input_opts) and ('depth_variable' not in input_opts):
        input_opts['depth_variable'] = 'depth'
    if ('depth_value' in input_opts) and ('depth_dimension' not in input_opts):
        input_opts['depth_dimension'] = 'depth'
    if 'variables' not in input_opts:
        input_opts['variables'] = 'temperature,salinity,u,v'
    if 'spatial_resolution' not in input_opts:
        input_opts['spatial_resolution'] = 12500
    if 'projection' not in input_opts:
        f_handler = netCDF4.Dataset(input_path, 'r')
        _crs = f_handler.variables['stereographic']
        crs = pyproj.CRS.from_cf({x: getattr(_crs, x) for x in _crs.ncattrs()})
        input_opts['projection'] = crs.to_wkt()
        f_handler.close()

    coords_converted = False
    grid_yx = idf_converter.readers.netcdf_grid_yx
    for result in grid_yx.read_data(input_opts, output_opts):
        input_opts, output_opts, granule, _transforms = result

        if not coords_converted:
            # x and y are provided as multiple of 100km, convert to meters...
            granule.vars['x']['array'] *= 100000
            granule.vars['y']['array'] *= 100000
            coords_converted = True

        if 'fice' in granule.vars.keys():
            transforms = []
            ice_mask = get_ice_mask(granule)

            ice_vars = [_ for _ in ('fice', 'hice', 'uice', 'vice')
                        if _ in granule.vars.keys()]
            mask_job = ('static_common_mask', {'targets': ice_vars,
                                               'mask': ice_mask})
            transforms.append(mask_job)

            transforms.extend(_transforms)
        else:
            transforms = _transforms

        yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2020-09-30
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module contains formatter methods for operations related to the
rasterization of vector features.
"""

import typing
import numpy
import numpy.typing
import PIL.Image
import PIL.ImageDraw
import shapely.geometry

GeoTransform = typing.Tuple[float, float, float, float, float, float]


def polygons_from_coords(linear_rings: typing.List[numpy.typing.NDArray],
                         geotransform: GeoTransform,
                         nrows: int,
                         ncols: int,
                         values: typing.Optional[typing.List[int]] = None
                         ) -> numpy.typing.NDArray:
    """"""
    ulx, dx, _, uly, _, dy = geotransform
    img_type = 'L'
    dtype = 'int'
    if values is None:
        img_type = '1'
        dtype = 'bool'
        values = [1] * len(linear_rings)
    img = PIL.Image.new(img_type, (nrows, ncols), 0)
    pencil = PIL.ImageDraw.Draw(img)
    for i, linear_ring in enumerate(linear_rings):
        coords = [((_[0] - ulx) / dx,
                   (_[1] - uly) / dy) for _ in linear_ring]
        pencil.polygon(coords, fill=values[i], outline=values[i])

    raster = numpy.array(img.getdata(0), dtype=dtype).reshape((ncols, nrows))
    img.close()

    return raster


def polygons_from_shapely(shapes: typing.List[shapely.geometry.Geometry],
                          geotransform: GeoTransform,
                          nrows: int,
                          ncols: int,
                          values: typing.Optional[typing.List[int]] = None
                          ) -> numpy.typing.NDArray:
    """"""
    ulx, dx, _, uly, _, dy = geotransform
    img_type = 'L'
    dtype = 'int'
    if values is None:
        img_type = '1'
        dtype = 'bool'
        values = [1] * len(shapes)
    img = PIL.Image.new(img_type, (nrows, ncols), 0)
    pencil = PIL.ImageDraw.Draw(img)
    for i, shape in enumerate(shapes):
        # print(values[i])
        if isinstance(shape, shapely.geometry.Polygon):
            coords = shape.exterior.coords
            coords = [((_[0] - ulx) / dx,
                       (_[1] - uly) / dy) for _ in coords]
            pencil.polygon(coords, fill=values[i], outline=values[i])
        elif isinstance(shape, shapely.geometry.MultiPolygon):
            for polygon in shape.geoms:
                coords = polygon.exterior.coords
                coords = [((_[0] - ulx) / dx,
                           (_[1] - uly) / dy) for _ in coords]
                pencil.polygon(coords, fill=values[i], outline=values[i])

    raster = numpy.array(img.getdata(0), dtype=dtype).reshape((ncols, nrows))
    img.close()

    return raster



# vim: ts=4:sts=4:sw=4
#
# @date 2020-12-20
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions
from idf_converter.lib.types import ReaderResult, TransformsList

logger = logging.getLogger(__name__)

DATA_MODEL = 'time_dependent'

flags = {'no_qc_performed': 0,
         'good_data': 1,
         'probably_good_data': 2,
         'bad_data_that_are_potentially_correctable': 3,
         'bad_data': 4,
         'value_changed': 5,
         'not_used': 6,
         'nominal_value': 7,
         'interpolated_value': 8,
         'missing_value': 9}


class InputPathMissing(ValueError):
    """Error raised when the input options have no "path" entry"""
    pass


class DepthValueMissing(ValueError):
    """Error raised when the input options have no "depth_value" entry."""
    pass


class DepthShapeNotSupported(Exception):
    """Error raised when the DEPH variable has an unsupported number of
    dimensions (0 or >2)"""
    pass


class NotEnoughData(Exception):
    """Error raised when the input data has less than 3 values, which is the
    minimal values count supported by this reader."""
    def __init__(self) -> None:
        """"""
        msg = 'This reader requires at least three along-track values'
        super(NotEnoughData, self).__init__(msg)


class InvalidDepthValue(Exception):
    """Error raised when the value passed as "depth_value" input option is not
    a valid float."""
    def __init__(self, str_value: str) -> None:
        """"""
        self.str_value = str_value
        msg = f'"{str_value}" cannot be converted into a numerical depth value'
        super(InvalidDepthValue, self).__init__(msg)


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    depth_value\tDepth of the drifter trajectory to convert',
           '    with-temperature\tRead temperature variable (yes)',
           '    with-day\tInclude day as a variable (yes)',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def mask_data(lon_mask: numpy.typing.NDArray, lat_mask: numpy.typing.NDArray,
              time_mask: numpy.typing.NDArray,
              pos_qc: numpy.typing.NDArray,
              time_qc: numpy.typing.NDArray
              ) -> numpy.typing.NDArray:
    list_bad_flags = ('bad_data_that_are_potentially_correctable', 'bad_data',
                      'missing_value')

    off_flags = numpy.uint32(0)
    for nflag in list_bad_flags:
        off_flags = off_flags + numpy.uint32(1 << flags[nflag])

    mask_on = (~lon_mask & ~ lat_mask & ~time_mask)

    # Valid mask = values are not masked and QC must have bad flags set to 0
    mask_on = (mask_on & (numpy.bitwise_and(time_qc, off_flags) == 0))
    mask_on = (mask_on & (numpy.bitwise_and(pos_qc, off_flags) == 0))
    result: numpy.typing.NDArray = numpy.ma.getdata(mask_on)
    return result


def mask_qual(qual: numpy.typing.NDArray
              ) -> numpy.typing.NDArray:
    list_bad_flags = ('bad_data_that_are_potentially_correctable', 'bad_data',
                      'missing_value')
    off_flags = numpy.uint32(0)
    for nflag in list_bad_flags:
        off_flags = off_flags + numpy.uint32(1 << flags[nflag])

    mask_off = numpy.zeros(numpy.shape(qual), dtype='bool')

    # Invalid mask = QC must have bad flags set to 1
    mask_off = (mask_off | (numpy.bitwise_and(qual, off_flags) > 0))

    result: numpy.typing.NDArray = numpy.ma.getdata(~mask_off)
    return result


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables = {'TEMP': {'name': 'temperature',
                          'long_name': 'Temperature',
                          'units': 'K',
                          'valid_min': 273.15,
                          'valid_max': 315.5,
                          'options': {}},
                 'EWCT': {'name': 'eastward_velocity',
                          'long_name': 'Eastward velocity',
                          'units': 'm.s-1',
                          'valid_min': -2.0,
                          'valid_max': 2.0,
                          'options': {}},
                 'NSCT': {'name': 'northward_velocity',
                          'long_name': 'Northward velocity',
                          'units': 'm',
                          'valid_min': -2.0,
                          'valid_max': 2.0,
                          'options': {}},
                 }
    dic_max = {}
    dic_min = {}
    # initialize valid min and max in a separate dictionary
    for var_id in variables.keys():
        dic_max[var_id] = variables[var_id]['valid_max']
        dic_min[var_id] = variables[var_id]['valid_min']
    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    # Read depth from input options
    _depth_value = input_opts.get('depth_value', None)
    if _depth_value is None:
        raise DepthValueMissing()

    try:
        depth_value = float(_depth_value)
    except ValueError:
        raise InvalidDepthValue(_depth_value)
    depth_str = f'{depth_value}'.rstrip('0').rstrip('.')

    _include_temp = input_opts.get('with-temperature', 'yes')
    include_temp = (_include_temp.lower() in ('yes', 'true', '1'))
    if include_temp is False:
        variables.pop('TEMP')

    _include_day = input_opts.get('with-day', 'yes')
    include_day = (_include_day.lower() in ('yes', 'true', '1'))

    # Define depth for all variables
    for var_id in variables.keys():
        variables[var_id]['depth'] = f'{depth_str}m'

    granule.vars = variables
    channels = list(variables.keys())

    lon_name = 'LONGITUDE'
    lat_name = 'LATITUDE'
    time_name = 'TIME'
    depth_name = 'DEPH'
    spatial_res_meters = 1.e07

    f_handler = netCDF4.Dataset(input_path, 'r')
    _time = idf_converter.lib.extract_variable_values(f_handler, time_name)
    _lon = idf_converter.lib.extract_variable_values(f_handler, lon_name)
    _lat = idf_converter.lib.extract_variable_values(f_handler, lat_name)
    _dep = idf_converter.lib.extract_variable_values(f_handler, depth_name)
    _pos_qc = f_handler.variables['POSITION_QC'][:]
    _time_qc = f_handler.variables[f'{time_name}_QC'][:]
    _tunits = f_handler.variables[time_name].units
    date = netCDF4.num2date(_time, units=_tunits)

    _lon_mask = numpy.ma.getmaskarray(_lon)
    if _lon_mask.any():
        logger.warning(f'Masked longitudes in {input_path}')
    _lon = numpy.ma.getdata(_lon)

    _lat_mask = numpy.ma.getmaskarray(_lat)
    if _lat_mask.any():
        logger.warning(f'Masked latitudes in {input_path}')
    _lat = numpy.ma.getdata(_lat)

    _time_mask = numpy.ma.getmaskarray(_time)
    if _time_mask.any():
        logger.warning(f'Masked times in {input_path}')
    _time = numpy.ma.getdata(_time)

    # mask non valid position / time
    _mask = mask_data(_lon_mask, _lat_mask, _time_mask, _pos_qc, _time_qc)
    if not _mask.any():
        logger.warning(f'No valid data in "{input_path}"')
        f_handler.close()
        raise idf_converter.lib.EarlyExit()

    # Only keep valid coordinates
    valid_ind = numpy.where(_mask)
    _time = _time[valid_ind] * 86400
    _lon = _lon[valid_ind]
    _lat = _lat[valid_ind]
    date = date[valid_ind]

    # Cut trajectory when geospatial or temporal information or missing
    dt = _time[1:] - _time[: -1]
    _ind = numpy.where(dt > 7 * 3600)[0]
    if _ind.any():
        for i in sorted(_ind, reverse=True):
            _lon = numpy.insert(_lon, i + 1, _lon[i + 1])
            _lon = numpy.insert(_lon, i + 1, _lon[i])
            _lat = numpy.insert(_lat, i + 1, _lat[i + 1])
            _lat = numpy.insert(_lat, i + 1, _lat[i])
            _time = numpy.insert(_time, i + 1, _time[i + 1] - 1)
            _time = numpy.insert(_time, i + 1, _time[i] + 1)
            date = numpy.insert(date, i + 1, date[i + 1])
            date = numpy.insert(date, i + 1, date[i])

    initial_size = _mask.shape[0]
    final_size = _time.shape[0]

    _dic: typing.Dict[str, numpy.typing.NDArray] = {}
    for var_id in granule.vars.keys():
        if var_id not in f_handler.variables.keys():
            tmp_data = numpy.full(final_size, numpy.nan, dtype=numpy.float32)
            granule.vars[var_id]['datatype'] = tmp_data.dtype
            _dic[var_id] = numpy.ma.masked_invalid(tmp_data)
            continue
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        qual_id = f'{var_id}_QC'
        qual = idf_converter.lib.extract_variable_values(f_handler, qual_id)

        _mask_qual = mask_qual(qual)
        band[~_mask_qual] = numpy.nan

        depth_ind = numpy.where(_dep == depth_value)
        if 0 >= len(depth_ind[0]):
            logger.warning(f'{var_id} not available at {depth_str}m depth '
                           f'in "{input_path}"')
            tmp_data = numpy.full(final_size, numpy.nan, dtype=numpy.float32)
            granule.vars[var_id]['datatype'] = tmp_data.dtype
            _dic[var_id] = numpy.ma.masked_invalid(tmp_data)
            continue

        if 2 == len(_dep.shape):
            # format before 2023-11-30: DEPH(TIME, DEPTH)
            _band1d = numpy.full(initial_size, numpy.nan, dtype=band.dtype)
            _band1d[depth_ind[0]] = band[depth_ind]
        elif 1 == len(_dep.shape):
            # format starting 2023-11-30: DEPH(DEPTH)
            _band1d = band[:, depth_ind]
        else:
            raise DepthShapeNotSupported()

        _band1d = _band1d[valid_ind]

        if _ind.any():
            for i in sorted(_ind, reverse=True):
                _band1d = numpy.insert(_band1d, i + 1, numpy.nan)
                _band1d = numpy.insert(_band1d, i + 1, numpy.nan)
        granule.vars[var_id]['datatype'] = band.dtype
        _dic[var_id] = numpy.ma.masked_invalid(_band1d)
    f_handler.close()

    # Compute Norm speed as SEAScope do not compute norm for trajectories yet
    nsct_mask = numpy.ma.getmaskarray(_dic['NSCT'])
    ewct_mask = numpy.ma.getmaskarray(_dic['EWCT'])
    _mask_finite = (~nsct_mask & ~ewct_mask)
    nsct_sqr: numpy.typing.NDArray = _dic['NSCT']**2
    ewct_sqr: numpy.typing.NDArray = _dic['EWCT']**2
    _speed = numpy.sqrt(nsct_sqr + ewct_sqr, where=_mask_finite)
    _speed[numpy.ma.getmaskarray(_speed)] = numpy.nan
    _dic['speed'] = numpy.ma.masked_invalid(_speed)

    max_speed = 0.0
    if not numpy.isnan(_dic['speed']).all():
        max_speed = numpy.nanmax(_dic['speed'])
    granule.vars['speed'] = {'name': 'speed',
                             'long_name': 'velocity norm',
                             'units': 'm/s',
                             'valid_min': 0,
                             'valid_max': max_speed,
                             'depth': f'{depth_str}m',
                             'datatype': granule.vars['NSCT']['datatype'],
                             'options': {}}
    channels.append('speed')
    dic_max['speed'] = max_speed
    dic_min['speed'] = 0

    # Translate Celsius into Kelvin
    if include_temp:
        valid_temp_ind = numpy.where(numpy.isfinite(_dic['TEMP']))
        if numpy.all((-10.0 < _dic['TEMP'][valid_temp_ind]) &
                     (100 > _dic['TEMP'][valid_temp_ind])):
            _dic['TEMP'][valid_temp_ind] = (273.15 +
                                            _dic['TEMP'][valid_temp_ind])

    # Add coordinates variables
    granule.vars['lat'] = {'units': 'degrees north',
                           'datatype': _lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'units': 'degrees east',
                           'datatype': _lon.dtype,
                           'options': {}
                           }
    _dic['lon'] = _lon
    _dic['lat'] = _lat

    # Apply time offset to transform time into an EPOCH timestamp
    ref_time = datetime.datetime(1950, 1, 1)
    epoch = datetime.datetime(1970, 1, 1)
    time_offset = (ref_time - epoch).total_seconds()
    shifted_time: numpy.typing.NDArray = _time + time_offset
    granule.vars['time'] = {'units': 'seconds since 1970-01-01T00:00:00.000Z',
                            'datatype': numpy.double,
                            'options': {}}
    _dic['time'] = shifted_time.astype(numpy.double)

    # Store day in month as SEAScope do not show position in time yet
    if include_day:
        # Store day in month as SEAScope do not show position in time yet
        dayinmonth = numpy.array([x.day for x in date])

        granule.vars['day'] = {'name': 'day_in_month',
                               'units': 'day in month',
                               'valid_min': 1,
                               'valid_max': 31,
                               'options': {}}
        _dic['day'] = dayinmonth
        channels.append('day')

    # Set Global parameters
    granule_name = os.path.splitext(os.path.basename(input_path))[0]
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = spatial_res_meters
    granule.meta['idf_spatial_resolution_units'] = 'm'

    # Split in monthly file to display only part of the trajectory in SEAScope
    list_yearmonth = [f'{x.year:04d}{x.month:02d}' for x in date]
    for ym in set(list_yearmonth):
        _ind_time = numpy.where(numpy.array(list_yearmonth) == f'{ym}')[0]
        # Get the monthly data plus the next index to have continuous
        # trajectories
        _slice = slice(_ind_time[0], _ind_time[-1] + 2)

        # ... unless there is actually a discontinuity
        if numpy.ma.getmaskarray(_dic['NSCT'])[_ind_time[-1]] is True:
            _slice = slice(_ind_time[0], _ind_time[-1] + 1)

        for var_id in granule.vars.keys():
            granule.vars[var_id]['array'] = _dic[var_id][_slice]
            if var_id in dic_min.keys():
                granule.vars[var_id]['valid_min'] = dic_min[var_id]
            if var_id in dic_max.keys():
                granule.vars[var_id]['valid_max'] = dic_max[var_id]

        # Trick to deal with continuity in longitude
        lonc: numpy.typing.NDArray = _dic['lon'][_slice]
        lonc_valid = lonc[numpy.where(numpy.isfinite(lonc))]
        lref = lonc_valid[int(0.5 * numpy.shape(lonc_valid)[0])]
        fixed_lon = numpy.mod(lonc - (lref - 180.), 360.) + (lref - 180.)
        granule.vars['lon']['array'] = fixed_lon

        granule.dims['time'] = numpy.size(granule.vars['time']['array'])
        _convertutc = datetime.datetime.utcfromtimestamp
        time_array: numpy.typing.NDArray = _dic['time'][_slice]
        start_dt = _convertutc(time_array[0])
        stop_dt = _convertutc(time_array[-1])
        if ym not in granule_name:
            _granule_name = f'{granule_name}_{ym}'
        else:
            _granule_name = granule_name
        granule.meta['idf_granule_id'] = f'{_granule_name}'
        granule.meta['time_coverage_start'] = start_dt
        granule.meta['time_coverage_end'] = stop_dt

        transforms: TransformsList = []

        transforms.append(('remove_extra_lon_degrees', {'lon_name': 'lon'}))

        output_opts['__export'] = channels

        # data_model, dims, vars, attrs, formatter_jobs
        yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-10
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import numpy
import numpy.typing
import typing
import netCDF4
import logging
import idf_converter.lib
import idf_converter.lib.time
from idf_converter.lib.types import Granule, ReaderResult, IdxSlices, IdxShifts
from idf_converter.lib.types import InputOptions, OutputOptions, TransformsList

logger = logging.getLogger(__name__)

ExtraValues = typing.Dict[str, typing.Dict[str, typing.Any]]
TimeValues = typing.Tuple[typing.Optional[str],
                          typing.Optional[numpy.typing.NDArray]]
DepthInfo = typing.Tuple[typing.Optional[str], slice]
TimeInfo = typing.Tuple[typing.Optional[str], slice]
DimInfo = typing.Tuple[str, bool, typing.Optional[int], typing.Optional[slice]]


class TimeCoverageNotFullyDefined(ValueError):
    """Error raised when the information available in the input file and in
    the input options do not provide enough information to deduce the time
    coverage."""
    pass


class DepthNotFound(Exception):
    """Error raised when the requested depth value is not available in the
    depth variable of the input file."""

    def __init__(self, value: float, tolerance: float) -> None:
        """"""
        self.value = value
        self.tolerance = tolerance


def get_depth_info(input_opts: InputOptions, granule: Granule,
                   idf_handler: netCDF4.Dataset) -> DepthInfo:
    """"""
    depth_dim_id = input_opts.get('depth_dimension', None)
    depth_var_id = input_opts.get('depth_variable', None)

    depth = None
    depth_attrs = {}
    depth_slice = slice(None, None, None)
    if depth_var_id is not None:
        _depth_var = idf_handler.variables[depth_var_id]
        depth = idf_converter.lib.extract_variable_values(idf_handler,
                                                          depth_var_id)
        for attr in _depth_var.ncattrs():
            depth_attrs[attr] = _depth_var.getncattr(attr)

        ndepth = depth.size

        _depth_value = input_opts.get('depth_value', None)
        if _depth_value is not None:
            depth_value = float(_depth_value)
            depth_ind = numpy.where(depth == depth_value)
            if 0 >= len(depth_ind[0]):
                # Look for the closest depth value which sits within the
                # tolerance range
                _depth_tolerance = input_opts.get('depth_tolerance', '0.0')
                depth_tolerance = float(_depth_tolerance)

                closest_ind = numpy.abs(depth - depth_value).argmin()
                closest_value = depth[closest_ind]
                depth_dist = numpy.abs(closest_value - depth_value)

                if depth_tolerance < depth_dist:
                    raise DepthNotFound(depth_value, depth_tolerance)

                depth_value = closest_value
                depth_ind = numpy.where(depth == depth_value)

            depth_slice = slice(depth_ind[0][0], depth_ind[0][0] + 1, None)
            ndepth = 1
            depth = numpy.array([depth_value])

        granule.vars['depth'] = depth_attrs
        granule.vars['depth']['array'] = depth
        granule.vars['depth']['options'] = {}
        granule.dims['depth'] = ndepth

    return (depth_dim_id, depth_slice)


def build_slices(ndims: int,
                 time_axis: int,
                 depth_axis: int,
                 dim1_axis: int,
                 dim2_axis: int,
                 time_slice: typing.Optional[slice],
                 depth_slice: typing.Optional[slice],
                 dim1_slice: typing.Optional[slice],
                 dim2_slice: typing.Optional[slice],
                 dim1_reversed: bool,
                 dim2_reversed: bool
                 ) -> IdxSlices:
    """"""
    # Define slices to extract and implement reverse traversal when
    # necessary
    axis_to_reverse = []
    if dim1_reversed:
        axis_to_reverse.append(dim1_axis)
    if dim2_reversed:
        axis_to_reverse.append(dim2_axis)
    _idx: IdxSlices = [slice(None, None, -1) if i in axis_to_reverse
                       else slice(None, None, None)
                       for i in range(0, ndims)]

    # Override time slice
    if (0 <= time_axis) and (time_slice is not None):
        _idx[time_axis] = time_slice

    # Override depth slice
    if (0 <= depth_axis) and (depth_slice is not None):
        _idx[depth_axis] = depth_slice

    # Override dim1 slice
    if (0 <= dim1_axis) and (dim1_slice is not None):
        _idx[dim1_axis] = dim1_slice

    # Override dim2 slice
    if (0 <= dim2_axis) and (dim2_slice is not None):
        _idx[dim2_axis] = dim2_slice

    return _idx


def get_slices_from_values(var_id: str,
                           extra_values: ExtraValues,
                           time_info: TimeInfo, depth_info: DepthInfo,
                           dim1_info: DimInfo, dim2_info: DimInfo
                           ) -> typing.Tuple[IdxSlices, IdxShifts]:
    """"""
    depth_dim_id, depth_slice = depth_info
    time_dim_id, time_slice = time_info
    dim1_id, dim1_reversed, dim1_shift, dim1_slice = dim1_info
    dim2_id, dim2_reversed, dim2_shift, dim2_slice = dim2_info

    dim1_axis = extra_values[var_id]['dims'].index(dim1_id)
    dim2_axis = extra_values[var_id]['dims'].index(dim2_id)
    depth_axis = -1
    if depth_dim_id is not None:
        if depth_dim_id in extra_values[var_id]['dims']:
            depth_axis = extra_values[var_id]['dims'].index(depth_dim_id)
    time_axis = -1
    if time_dim_id is not None:
        if time_dim_id in extra_values[var_id]['dims']:
            time_axis = extra_values[var_id]['dims'].index(time_dim_id)

    ndims = len(extra_values[var_id]['dims'])

    idx = build_slices(ndims, time_axis, depth_axis, dim1_axis, dim2_axis,
                       time_slice, depth_slice, dim1_slice, dim2_slice,
                       dim1_reversed, dim2_reversed)

    shift_info = {}
    if dim1_shift is not None:
        shift_info[dim1_axis] = dim1_shift
    if dim2_shift is not None:
        shift_info[dim2_axis] = dim2_shift

    return idx, shift_info


def get_slices_from_netcdf(var_id: str, f_handler: netCDF4.Dataset,
                           time_info: TimeInfo, depth_info: DepthInfo,
                           dim1_info: DimInfo, dim2_info: DimInfo
                           ) -> typing.Tuple[IdxSlices, IdxShifts]:
    """"""
    depth_dim_id, depth_slice = depth_info
    time_dim_id, time_slice = time_info
    dim1_id, dim1_reversed, dim1_shift, dim1_slice = dim1_info
    dim2_id, dim2_reversed, dim2_shift, dim2_slice = dim2_info

    netcdf_var = f_handler.variables[var_id]
    dim1_axis: int = netcdf_var.dimensions.index(dim1_id)
    dim2_axis: int = netcdf_var.dimensions.index(dim2_id)
    depth_axis = -1
    if depth_dim_id is not None:
        if depth_dim_id in netcdf_var.dimensions:
            depth_axis = netcdf_var.dimensions.index(depth_dim_id)
    time_axis = -1
    if time_dim_id is not None:
        if time_dim_id in netcdf_var.dimensions:
            time_axis = netcdf_var.dimensions.index(time_dim_id)

    ndims = len(f_handler.variables[var_id].dimensions)

    idx = build_slices(ndims, time_axis, depth_axis, dim1_axis, dim2_axis,
                       time_slice, depth_slice, dim1_slice, dim2_slice,
                       dim1_reversed, dim2_reversed)

    shift_info = {}
    if dim1_shift is not None:
        shift_info[dim1_axis] = dim1_shift
    if dim2_shift is not None:
        shift_info[dim2_axis] = dim2_shift

    return idx, shift_info


def extract_values(granule: Granule, f_handler: netCDF4.Dataset,
                   extra_values: ExtraValues,
                   time_info: DepthInfo, depth_info: DepthInfo,
                   dim1_info: DimInfo,
                   dim2_info: DimInfo,
                   var_id: str) -> None:
    """"""
    depth_dim_id, depth_slice = depth_info
    time_dim_id, time_slice = time_info

    band: numpy.typing.NDArray
    if var_id in extra_values:
        idx, shift_info = get_slices_from_values(var_id, extra_values,
                                                 time_info, depth_info,
                                                 dim1_info, dim2_info)

        # Define basic information about the variable
        var_units = extra_values[var_id].get('units', '')
        granule.vars[var_id] = {'options': {}, 'name': var_id,
                                'units': var_units}

        _idx = tuple(idx)
        if 0 >= len(shift_info):
            band = extra_values[var_id]['array'][_idx]
        else:
            band = extra_values[var_id]['array']
            for axis, amount in shift_info.items():
                band = numpy.roll(band, amount, axis=axis)
            band = band[_idx]
    else:
        idx, shift_info = get_slices_from_netcdf(var_id, f_handler,
                                                 time_info, depth_info,
                                                 dim1_info, dim2_info)

        # Extract as much information as possible from the input variable
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)

        band = idf_converter.lib.extract_variable_values(f_handler, var_id,
                                                         None, idx, shift_info)

    flag_var_id = granule.vars[var_id].get('flag_variable', None)
    if flag_var_id is not None:

        _idx = tuple(idx)

        # Read flag variable
        flag_band: numpy.typing.NDArray
        if flag_var_id in extra_values:
            if 0 >= len(shift_info):
                flag_band = extra_values[flag_var_id]['array'][_idx]
            else:
                flag_band = extra_values[flag_var_id]['array']
                for axis, amount in shift_info.items():
                    flag_band = numpy.roll(flag_band, amount, axis=axis)
                flag_band = flag_band[_idx]
        else:
            if 0 >= len(shift_info):
                flag_band = f_handler.variables[flag_var_id][_idx]
            else:
                flag_band = f_handler.variables[flag_var_id]
                for axis, amount in shift_info.items():
                    flag_band = numpy.roll(flag_band, amount, axis=axis)
                flag_band = flag_band[_idx]

        flag_mask = idf_converter.lib.build_flag_mask(var_id, band, flag_band,
                                                      granule)
        # Update band mask
        band = numpy.ma.masked_where(flag_mask, band)

    if isinstance(band, numpy.ma.MaskedArray):
        granule.vars[var_id]['array'] = band
    else:
        granule.vars[var_id]['array'] = numpy.ma.array(band)

    has_valid_min = 'valid_min' in granule.vars[var_id].keys()
    has_valid_max = 'valid_max' in granule.vars[var_id].keys()
    incomplete_valid_range = False in (has_valid_min, has_valid_max)

    auto_min = (has_valid_min and
                ('auto' == granule.vars[var_id]['valid_min']))
    auto_max = (has_valid_max and
                ('auto' == granule.vars[var_id]['valid_max']))
    has_undefined_extrema = True in (auto_min, auto_max)

    # Compute min/max from values is valid_min/valid_max attributes have
    # not been provided
    band_mask = numpy.ma.getmaskarray(band)
    if band_mask.all():
        logger.warning(f'All values are masked for variable {var_id}')
        if (has_valid_min is False) or auto_min:
            logger.warning('valid_min not specified and no way to guess it '
                           'from input file: please provide it using the '
                           'variable_override mechanism')
        if (has_valid_max is False) or auto_max:
            logger.warning('valid_max not specified and no way to guess it '
                           'from input file: please provide it using the '
                           'variable_override mechanism')
        if incomplete_valid_range or has_undefined_extrema:
            logger.warning(f'Using default [0, 1] valid range for {var_id}')
            granule.vars[var_id]['valid_min'] = band.dtype.type(0)
            granule.vars[var_id]['valid_max'] = band.dtype.type(1)
    else:
        finite_ind = numpy.where(numpy.isfinite(band))
        if (has_valid_min is False) or auto_min:
            granule.vars[var_id]['valid_min'] = numpy.min(band[finite_ind])
        if (has_valid_max is False) or auto_max:
            granule.vars[var_id]['valid_max'] = numpy.max(band[finite_ind])


def single_granule(granule: Granule,
                   f_handler: netCDF4.Dataset,
                   var_ids: typing.List[str],
                   depth_info: DepthInfo,
                   dim1_info: DimInfo,
                   dim2_info: DimInfo,
                   input_opts: InputOptions,
                   output_opts: OutputOptions,
                   time_dim_id: typing.Optional[str],
                   dtimes: numpy.typing.NDArray
                   ) -> typing.Iterator[ReaderResult]:
    """"""
    # Add transforms
    transforms: TransformsList = []

    if 'indicative_time' in input_opts:
        dt_str = input_opts['indicative_time']
        indicative_dt = idf_converter.lib.time.parse_datetime(dt_str)
        dtimes = numpy.array([indicative_dt])

    coverages = idf_converter.lib.time.get_time_coverages(input_opts, granule,
                                                          dtimes)
    if coverages[0] is None:
        logger.error('Impossible to create a granule without a complete '
                     'time coverage')
        raise TimeCoverageNotFullyDefined()

    start_dt, stop_dt = coverages[0]
    if start_dt == stop_dt:
        logger.warning('Time coverage has no duration, i.e. '
                       'time_coverage_start = time_coverage_end')
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt

    extra_values = input_opts.get('extra_values', {})
    sub_time_info = (time_dim_id, slice(None, None, None))
    for var_id in var_ids:
        extract_values(granule, f_handler, extra_values, sub_time_info,
                       depth_info, dim1_info, dim2_info, var_id)

    output_opts['__export'] = var_ids
    yield (input_opts, output_opts, granule, transforms)


def multi_granule(granule: Granule,
                  f_handler: netCDF4.Dataset,
                  var_ids: typing.List[str],
                  depth_info: DepthInfo,
                  dim1_info: DimInfo,
                  dim2_info: DimInfo,
                  input_opts: InputOptions,
                  output_opts: OutputOptions,
                  time_dim_id: typing.Optional[str],
                  dtimes: numpy.typing.NDArray
                  ) -> typing.Iterator[ReaderResult]:
    """"""
    coverages = idf_converter.lib.time.get_time_coverages(input_opts, granule,
                                                          dtimes)

    extra_values = input_opts.get('extra_values', {})
    for time_ind, dtime in enumerate(dtimes):
        if (dtime is None) or (dtime is numpy.ma.masked):
            logger.warning(f'{time_ind+1}th time is invalid and has been '
                           'skipped')
            continue

        start_dt, stop_dt = coverages[time_ind]
        logger.debug(f'Processing sub-granule {dtime:%Y-%m-%dT%H:%M:%SZ}')

        if (start_dt is None) or (stop_dt is None):
            raise Exception('')  # TODO

        subgranule = idf_converter.lib.create_subgranule(granule,
                                                         output_opts,
                                                         start_dt, stop_dt,
                                                         dtime)
        sub_time_info = (time_dim_id, slice(time_ind, time_ind + 1, None))
        for var_id in granule.vars.keys():
            logger.debug(f'\textracting data for variable {var_id}')
            subgranule.vars[var_id] = {}
            for attr in granule.vars[var_id].keys():
                # Ancillary variables should already have their "array"
                # attribute (lon, lat, etc...) so it will be copied to the
                # subgranule, but other variables should not have this
                # attribute yet
                subgranule.vars[var_id][attr] = granule.vars[var_id][attr]

        for var_id in var_ids:
            extract_values(subgranule, f_handler, extra_values,
                           sub_time_info, depth_info, dim1_info,
                           dim2_info, var_id)

        output_opts['__export'] = var_ids.copy()

        # Add transforms
        transforms: TransformsList = []

        yield (input_opts, output_opts, subgranule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import typing
import logging

import numpy
import numpy.typing

from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)


class UnknownPolarisation(Exception):
    def __init__(self, pol: str):
        self.polarisation = pol


def cmodhh(u10: numpy.typing.NDArray,
           phi: numpy.typing.NDArray,
           inc: numpy.typing.NDArray
           ) -> numpy.typing.NDArray:
    """
    """
    # Zhang Biao, Mouche Alexis, Lu Yiru, Perrie William, Zhang Guosheng,
    # Wang He (2019).
    # A Geophysical Model Function for Wind Speed Retrieval from C-Band
    # HH-polarized Synthetic Aperture Radar.
    # Ieee Geoscience And Remote Sensing Letters, 16(10), 1521-1525.
    # Publisher's official version : https://doi.org/10.1109/LGRS.2019.2905578
    # Open Access version : https://archimer.ifremer.fr/doc/00485/59715/

    # Coefficients and constants
    c = [
        0.,  # fake coeff in order to start with c[1]
        -0.72722756511, -1.1901195406, 0.33968637656, 0.086759069544,
        0.003090124916, 0.011761378188, 0.129158495658, 0.083506931034,
        4.092557781322, 1.211169044551, -1.119776245438, 0.579066509504,
        -0.604527699539, 0.118371042255, 0.008955505675, 0.219608674529,
        0.017557536680, 24.442309754388, 1.983490330585, 6.781440647278,
        7.947947040974, -4.696499003167, -0.437054238710, 5.471252046908,
        0.639468224273, 0.673385731705, 3.433229044819, 0.367036215316
    ]
    zpow = 1.6
    thetm = 40.
    thethr = 25.
    y0 = c[19]
    pn = c[20]
    a = y0 - (y0 - 1.) / pn
    b = 1. / (pn * (y0 - 1.) ** (pn - 1.))

    # Angles
    cosphi = numpy.cos(numpy.deg2rad(phi))
    x = (inc - thetm) / thethr
    x2 = x ** 2.

    # B0 term
    a0 = c[1] + c[2] * x + c[3] * x2 + c[4] * x * x2
    a1 = c[5] + c[6] * x
    a2 = c[7] + c[8] * x
    gam = c[9] + c[10] * x + c[11] * x2
    s0 = c[12] + c[13] * x
    s = a2 * u10
    del a2
    a3 = 1. / (1. + numpy.exp(-s0))
    slts0 = s < s0
    a3[~slts0] = 1. / (1. + numpy.exp(-s[~slts0]))
    a3_pow = (s0[slts0] * (1. - a3[slts0]))
    a3[slts0] = a3[slts0] * (s[slts0] / s0[slts0]) ** a3_pow
    del a3_pow
    del s
    del s0
    del slts0
    b0 = (a3 ** gam) * 10. ** (a0 + a1 * u10)
    del gam
    del a0
    del a1
    del a3

    # B1 term
    b1 = c[15] * u10 * (0.5 + x - numpy.tanh(4. * (x + c[16] + c[17] * u10)))
    b1 = (c[14] * (1. + x) - b1) / (numpy.exp(0.34 * (u10 - c[18])) + 1.)

    # B2 term
    v0 = c[21] + c[22] * x + c[23] * x2
    d1 = c[24] + c[25] * x + c[26] * x2
    d2 = c[27] + c[28] * x
    del x
    del x2
    v2 = (u10 / v0 + 1.)
    del v0
    v2lty0 = v2 < y0
    v2[v2lty0] = a + b * (v2[v2lty0] - 1.) ** pn
    b2 = (-d1 + d2 * v2) * numpy.exp(-v2)
    del d1
    del d2
    del v2

    # Sigma0 according to Fourier terms
    sig = (b0 * (1. + b1 * cosphi + b2 * (2. * cosphi ** 2. - 1.))) ** zpow
    del b0
    del b1
    del b2
    del cosphi
    return sig


def cmod5(u10: numpy.typing.NDArray,
          phi: numpy.typing.NDArray,
          inc: numpy.typing.NDArray,
          neutral: bool = True
          ) -> numpy.typing.NDArray:
    """
    """
    # Coefficients and constants
    if neutral is True:  # CMOD5.n coefficients
        c = [0., -0.6878, -0.7957, 0.338, -0.1728, 0., 0.004, 0.1103,
             0.0159, 6.7329, 2.7713, -2.2885, 0.4971, -0.725, 0.045, 0.0066,
             0.3222, 0.012, 22.7, 2.0813, 3., 8.3659, -3.3428, 1.3236,
             6.2437, 2.3893, 0.3249, 4.159, 1.693]
    else:  # CMOD5 coefficients
        c = [0., -0.688, -0.793, 0.338, -0.173, 0., 0.004, 0.111,
             0.0162, 6.34, 2.57, -2.18, 0.4, -0.6, 0.045, 0.007,
             0.33, 0.012, 22., 1.95, 3., 8.39, -3.44, 1.36, 5.35,
             1.99, 0.29, 3.80, 1.53]
    zpow = 1.6
    thetm = 40.
    thethr = 25.
    y0 = c[19]
    pn = c[20]
    a = y0 - (y0 - 1.) / pn
    b = 1. / (pn * (y0 - 1.) ** (pn - 1.))

    # Angles
    cosphi = numpy.cos(numpy.deg2rad(phi))
    x = (inc - thetm) / thethr
    x2 = x ** 2.

    # B0 term
    a0 = c[1] + c[2] * x + c[3] * x2 + c[4] * x * x2
    a1 = c[5] + c[6] * x
    a2 = c[7] + c[8] * x
    gam = c[9] + c[10] * x + c[11] * x2
    s0 = c[12] + c[13] * x
    s = a2 * u10
    del a2
    a3 = 1. / (1. + numpy.exp(-s0))
    slts0 = s < s0
    a3[~slts0] = 1. / (1. + numpy.exp(-s[~slts0]))
    a3_pow = (s0[slts0] * (1. - a3[slts0]))
    a3[slts0] = a3[slts0] * (s[slts0] / s0[slts0]) ** a3_pow
    del a3_pow
    del s
    del s0
    del slts0
    b0 = (a3 ** gam) * 10. ** (a0 + a1 * u10)
    del gam
    del a0
    del a1
    del a3

    # B1 term
    b1 = c[15] * u10 * (0.5 + x - numpy.tanh(4. * (x + c[16] + c[17] * u10)))
    b1 = (c[14] * (1. + x) - b1) / (numpy.exp(0.34 * (u10 - c[18])) + 1.)

    # B2 term
    v0 = c[21] + c[22] * x + c[23] * x2
    d1 = c[24] + c[25] * x + c[26] * x2
    d2 = c[27] + c[28] * x
    del x
    del x2
    v2 = (u10 / v0 + 1.)
    del v0
    v2lty0 = v2 < y0
    v2[v2lty0] = a + b * (v2[v2lty0] - 1.) ** pn
    b2 = (-d1 + d2 * v2) * numpy.exp(-v2)
    del d1
    del d2
    del v2

    # Sigma0 according to Fourier terms
    sig = b0 * (1. + b1 * cosphi + b2 * (2. * cosphi ** 2. - 1.)) ** zpow
    del b0
    del b1
    del b2
    del cosphi
    return sig


def _compute_roughness(sigma0: numpy.typing.NDArray,
                       incidence: numpy.typing.NDArray,
                       polarisation: str
                       ) -> numpy.typing.NDArray:
    """Compute sea surface roughness.

    Parameters
    ----------
    sigma0 : ndarray
        NRCS backscatter.
    incidence : ndarray
        Incidence angle in degrees.
    polarisation : str
        'VV' or 'HH' or 'VH' or 'HV'

    Returns
    -------
    ndarray
    """
    if polarisation == 'VV':  # Use cmod5
        sigma0_vv = cmod5(10., 45., incidence)
        return sigma0 / sigma0_vv
    elif polarisation == 'HH':  # Use cmodhh
        sigma0_hh = cmodhh(10., 45., incidence)
        return sigma0 / sigma0_hh
    elif polarisation == 'VH' or polarisation == 'HV':  # Use simple model
        # nrcs_vh_db = 0.580*wsp - 35.652
        # nrcs_vh_lin = 10^(nrcs_vh_db/10.)
        sigma0_cross = 10 ** ((0.58 * 10. - 35.652) / 10.)
        return sigma0 / sigma0_cross
    else:
        raise UnknownPolarisation(polarisation)


def compute_roughness(input_opts: InputOptions,
                      output_opts: OutputOptions,
                      granule: Granule,
                      *args: typing.Iterable[typing.Any],
                      targets: typing.Sequence[str],
                      sigma0_var_id: str,
                      incidence_var_id: str,
                      polarisation: str) -> FormatterResult:
    """"""
    # Inputs
    sigma0 = granule.vars[sigma0_var_id]['array']
    incidence = granule.vars[incidence_var_id]['array']
    # Roughness calculation
    # By blocks in order to decrease memory usage
    naz, nra = sigma0.shape
    roughness = numpy.empty((naz, nra), dtype='float32')
    blocksize = 25000000
    block_nra = nra
    block_naz = numpy.maximum(blocksize // block_nra, 1)
    nblock = int(numpy.ceil(naz / float(block_naz)))
    for iblock in range(nblock):
        az0 = iblock * block_naz
        az1 = numpy.minimum((iblock + 1) * block_naz, naz)
        az_slice = slice(az0, az1)
        _roughness = _compute_roughness(sigma0[az_slice, :],
                                        incidence[az_slice, :],
                                        polarisation)
        roughness[az_slice, :] = numpy.sqrt(_roughness)
    granule.vars[targets[0]]['array'] = roughness

    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4

import os
import sys
import shutil
import typing
import logging
from setuptools import setup

logger = logging.getLogger()
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)
logger.setLevel(logging.WARN)


# Check Python version
if sys.version_info < (3, 6):
    logger.error('This package is only available for Python version >= 3.6')
    sys.exit(1)

__package_name__ = 'idf_converter'

project_dir = os.path.dirname(__file__)
git_exe = shutil.which('git')
git_dir = os.path.join(project_dir, '.git')
version_file = os.path.join(project_dir, 'VERSION.txt')
readme_file = os.path.join(project_dir, 'docs', 'README.rst')
package_dir = os.path.join(project_dir, __package_name__)
init_file = os.path.join(package_dir, '__init__.py')

# Read metadata from the main __init__.py file
metadata: typing.Dict[str, typing.Any] = {}
with open(init_file, 'rt') as f:
    exec(f.read(), metadata)

# Regenerate a version file from git history
if git_exe is not None and os.path.isdir(git_dir):
    import subprocess
    gitrev = (git_exe, 'rev-list', 'HEAD', '--count')
    major, minor = metadata['__version__'].split('.')
    commits = subprocess.check_output(gitrev).decode('utf-8').strip()
    with open(version_file, 'wt') as f:
        f.write(f'{major}.{minor}.{commits}')

# Refuse to install package if version is not available
if not os.path.isfile(version_file):
    logger.error(f'Version file "{version_file}" missing.')
    logger.error('Please use a proper release of the code.')
    sys.exit(1)

with open(version_file, 'rt') as f:
    version = f.read()

with open(readme_file, 'rt') as f:
    long_description = f.read()

requirements: typing.List[str] = []
with open('requirements.txt', 'r') as f:
    lines = [x.strip() for x in f if 0 < len(x.strip())]
    requirements = [x for x in lines if x[0].isalpha()]

_pkg = __package_name__
cmds = [f'idf-converter = {_pkg}.cli:idf_converter_script']
readers = [f'sentinel3/sral/L2 = {_pkg}.readers.sentinel3_sral_l2',
           f'sentinel3/olci/L1 = {_pkg}.readers.sentinel3_olci_l1',
           f'sentinel3/olci/L2 = {_pkg}.readers.sentinel3_olci_l2',
           f'sentinel3/slstr/L1/bt = {_pkg}.readers.sentinel3_slstr_l1_bt',
           f'sentinel3/slstr/L1/rad = {_pkg}.readers.sentinel3_slstr_l1_rad',
           f'sentinel3/slstr/L2 = {_pkg}.readers.sentinel3_slstr_l2',
           f'sentinel6/p4/L2 = {_pkg}.readers.sentinel6_p4_l2',
           f'netcdf/grid/latlon = {_pkg}.readers.netcdf_grid_latlon',
           f'netcdf/grid/yx = {_pkg}.readers.netcdf_grid_yx',
           f'bremen/amsr/L3/sic = {_pkg}.readers.bremen_amsr_l3_sic',
           f'remss/L3/wind/netcdf = {_pkg}.readers.remss_l3_wind_netcdf',
           f'remss/L3/wind/bytemap = {_pkg}.readers.remss_l3_wind_bytemap',
           f'model/mfwam = {_pkg}.readers.mfwam',
           f'model/topaz4 = {_pkg}.readers.topaz4',
           f'ifremer/scat/psi-backscatter = '
           f'{_pkg}.readers.ifremer_scat_psi_backscatter',
           f'smos/miras/L2 = {_pkg}.readers.smos_miras_l2',
           f'jamstec/argo_oi/monthly = {_pkg}.readers.jamstec_argo_oi_monthly',
           f'iprc/argo_vi/monthly = {_pkg}.readers.iprc_argo_vi_monthly',
           f'roemmich-gilson/monthly = {_pkg}.readers.roemmich_gilson_monthly',
           f'metop/ascat/L2/wind = {_pkg}.readers.metop_ascat_l2_wind',
           f'metoffice/en4/analysis = {_pkg}.readers.metoffice_en4_analysis',
           f'isas/monthly = {_pkg}.readers.isas_monthly',
           f'noaa/cmorph/3-hourly = {_pkg}.readers.cmorph',
           f'woa/monthly = {_pkg}.readers.woa_monthly',
           f'sentinel1/L2/owi = {_pkg}.readers.sentinel1_l2_owi',
           f'sentinel1/L2/rvl = {_pkg}.readers.sentinel1_l2_rvl',
           f'ghrsst/l2p = {_pkg}.readers.ghrsst_l2p',
           f'osisaf/L3/sic = {_pkg}.readers.osisaf_l3_sic',
           f'cmems/drifter = {_pkg}.readers.cmems_drifter',
           f'rtofs = {_pkg}.readers.rtofs',
           f'ibtracs = {_pkg}.readers.ibtracs',
           f'sentinel1/L1 = {_pkg}.readers.sentinel1_l1',
           f'sentinel2/L1 = {_pkg}.readers.sentinel2_l1c',
           f'swot/L3 = {_pkg}.readers.swot_l3',
           ]
fmts = [f'remove_extra_lon_degrees = {_pkg}.lib.geo:remove_extra_lon_degrees',
        f'fill_1d_gaps = {_pkg}.lib.interp:fill_1d_gaps',
        f'static_common_mask = {_pkg}.lib.mask:static_common_mask',
        f'static_bands_mask = {_pkg}.lib.mask:static_bands_mask',
        f'mask_methods = {_pkg}.lib.mask:mask_methods',
        f'share_masks = {_pkg}.lib.mask:share_masks',
        f'store_as_ubytes = {_pkg}.lib.data_type:as_ubytes',
        f'store_as_float32 = {_pkg}.lib.data_type:as_float32',
        f's3_cnes_rayleigh = {_pkg}.lib.atmos_correction:s3_cnes_rayleigh',
        f'contrast_from_pct = {_pkg}.lib.scale:contrast_from_pct',
        f'stretch_extent = {_pkg}.lib.scale:stretch_extent',
        f'logscale = {_pkg}.lib.scale:logscale',
        f'limit_extrema = {_pkg}.lib.scale:limit_extrema',
        f'share_extrema = {_pkg}.lib.scale:share_extrema',
        f'extrema_methods = {_pkg}.lib.scale:extrema_methods',
        f'median_filter = {_pkg}.lib.filter:median',
        f'n2_sst_from_slstr = {_pkg}.lib.n2_sst:n2_sst_from_slstr',
        f's3_nir_bc = {_pkg}.lib.nir_bc:brightness_contrast',
        f'dir2vectors = {_pkg}.lib.vectors:from_dir',
        f'remove_vars = {_pkg}.lib.granule:remove_vars',
        f's1_experimental_radial_velocity = {_pkg}.lib.doppler:s1_radvel_exp',
        f'radial_velocity_sign = {_pkg}.lib.doppler:radial_velocity_sign',
        f'anomaly_from_clim = {_pkg}.lib.anomaly:anomaly_from_clim',
        f'compute_roughness = {_pkg}.lib.sar:compute_roughness',
        f'compute_swot_roughness = {_pkg}.lib.swot:compute_swot_roughness',
        f'compute_swot_ssha_unedited_nolr = {_pkg}.lib.swot:compute_swot_ssha_unedited_nolr',  # noqa
        ]
setup(
    zip_safe=False,
    name=__package_name__,
    version=version,
    author=metadata['__author__'],
    author_email=metadata['__author_email__'],
    url=metadata['__url__'],
    keywords=metadata['__keywords__'],
    classifiers=metadata['__classifiers__'],
    description=metadata['__description__'],
    packages=(__package_name__,
              f'{__package_name__}.lib',
              f'{__package_name__}.lib.idf',
              f'{__package_name__}.lib.idf.v1',
              f'{__package_name__}.readers'),
    license='COPYING.LESSER',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    package_data={__package_name__: ['py.typed', 'share/*.*']},
    test_suite='tests',
    python_requires='>=3.7',
    install_requires=requirements,
    setup_requires=(),
    entry_points={'console_scripts': cmds, 'idf.input.readers': readers,
                  'idf.formatters': fmts},
)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module provides methods to format granules according to version 1.x of
the IDF format."""

import os
import numpy
import typing
import logging
import netCDF4
import datetime
import idf_converter
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ProcessedLists
from idf_converter.lib.types import Granule, FormatterResult, TransformsList
try:
    import importlib.metadata
    idf_converter_version = importlib.metadata.version('idf_converter')
except ImportError:
    import pkg_resources
    _idf_converter_pkg = pkg_resources.get_distribution('idf_converter')
    idf_converter_version = _idf_converter_pkg.version

logger = logging.getLogger(__name__)
Granule2IDFResult = typing.Optional[ProcessedLists]
WriteVariableResult = typing.Tuple[netCDF4.Variable, typing.List[str]]


class MissingInputPath(ValueError):
    """Error raised if the "path" key is missing from the input options."""
    pass


class MissingOutputPath(ValueError):
    """Error raised if the "path" key is missing from the output options."""
    pass


class MissingCollection(ValueError):
    """Error raised if the "collection" key is missing from the output
    options."""
    pass


class MissingExportList(ValueError):
    """Error raised if the "__export" key is either missing or associated with
    an empty list in the output options."""
    pass


class DataModelNotSupported(ValueError):
    """Error raised when the granule to convert has a data model which is not
    supported by this version of IDF.

    Parameters
    ----------
    data_model: str
        Data model associated with the granule to convert and which is not
        supported by this version of the IDF format
    """
    def __init__(self, data_model: str) -> None:
        """"""
        self.data_model = data_model
        super(DataModelNotSupported, self).__init__()


def help() -> typing.Tuple[str, str]:
    """"""
    inp = ('',)  # "path" is required but will also be claimed by readers
    out = ('    path\tPath of the directory where the IDF files will be saved',
           '    collection\tName of the collection that will own the IDF files'
           ' produced by the converter',
           )
    return ('\n'.join(inp), '\n'.join(out))


def apply_data_storage_policy(input_opts: InputOptions,
                              output_opts: OutputOptions,
                              granule: Granule,
                              transforms: TransformsList
                              ) -> None:
    """"""
    def _lazy_eval_ubyte_export(output_opts: OutputOptions,
                                granule: Granule
                                ) -> typing.Iterator[str]:
        output_var_ids = output_opts.get('__export', None)
        if (output_var_ids is None) or (0 >= len(output_var_ids)):
            raise MissingExportList()

        _float_variables = output_opts.get('use_float32_values', '')
        float_variables = [_.strip() for _ in _float_variables.split(',')
                           if 0 < len(_.strip())]
        var_names = {_: granule.vars[_]['name'] for _ in output_var_ids}
        for var_id in output_var_ids:
            if var_names[var_id] not in float_variables:
                yield var_id

    def _lazy_eval_float_export(output_opts: OutputOptions,
                                granule: Granule
                                ) -> typing.Iterator[str]:
        output_var_ids = output_opts.get('__export', None)
        if (output_var_ids is None) or (0 >= len(output_var_ids)):
            raise MissingExportList()

        _float_variables = output_opts.get('use_float32_values', '')
        float_variables = [_.strip() for _ in _float_variables.split(',')
                           if 0 < len(_.strip())]
        var_names = {_: granule.vars[_]['name'] for _ in output_var_ids}
        for var_id in output_var_ids:
            if var_names[var_id] in float_variables:
                yield var_id

    # PACKING AS UBYTES
    # It automatically sets NaN and masked values to fill value (255)
    transforms.append(('store_as_ubytes',
                       {'targets': _lazy_eval_ubyte_export(output_opts,
                                                           granule)}))

    # STORING AS FLOAT32
    transforms.append(('store_as_float32',
                       {'targets': _lazy_eval_float_export(output_opts,
                                                           granule)}))


def preformat(input_opts: InputOptions,
              output_opts: OutputOptions,
              granule: Granule) -> typing.Iterator[FormatterResult]:
    """"""
    data_model_mod = get_data_model(granule.data_model)
    subgranules = data_model_mod.preformat(input_opts, output_opts, granule)
    for subgranule in subgranules:
        original_id = subgranule[2].meta['idf_granule_id']
        granule_id_prefix = output_opts.get('id_prefix', '')
        if 'INPUT_FILENAME' == granule_id_prefix:
            input_filename = os.path.basename(input_opts['path'])
            granule_id_prefix, _ = os.path.splitext(input_filename)
            if granule_id_prefix.endswith('_') is False:
                granule_id_prefix = f'{granule_id_prefix}_'
        granule_id_suffix = output_opts.get('id_suffix', '')
        if 'INPUT_FILENAME' == granule_id_suffix:
            input_filename = os.path.basename(input_opts['path'])
            granule_id_suffix, _ = os.path.splitext(input_filename)
            if granule_id_suffix.startswith('_') is False:
                granule_id_suffix = f'_{granule_id_suffix}'
        new_id = f'{granule_id_prefix}{original_id}{granule_id_suffix}'
        subgranule[2].meta['idf_granule_id'] = new_id
        yield subgranule


def get_idf_path(input_opts: InputOptions,
                 output_opts: OutputOptions,
                 granule: Granule) -> typing.Tuple[str, str]:
    """Build the path of the granule directory and the name of the IDF file
    from the input and ouput options.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.types.Granule
        Object representing the granule data and metadata

    Returns
    -------
    tuple
        A tuple containing two elements:

        - the path of the granule directory
        - the name of the IDF file
    """
    output_dir = output_opts.get('path', None)
    if output_dir is None:
        raise MissingOutputPath()

    collection = output_opts.get('collection', None)
    if collection is None:
        raise MissingCollection()

    collection_dir = os.path.join(output_dir, collection)

    input_path = input_opts.get('path', None)
    if input_path is None:
        raise MissingInputPath()

    granule_name = granule.meta['idf_granule_id']
    granule_dir = os.path.join(collection_dir, granule_name)
    idf_name = f'{granule_name}_idf_00.nc'
    return (granule_dir, idf_name)


def _to_numerical(str_value: str) -> float:
    """"""
    try:
        result = float(str_value)
    except ValueError:
        try:
            result = float(''.join([c for c in str_value
                                    if c.isdigit() or '.' == c]))
        except ValueError:
            result = float('nan')
    return result


def _to_spec_time_format(time_obj: typing.Any, specs: typing.Any) -> str:
    """"""
    _time_obj = time_obj
    if not isinstance(_time_obj, datetime.datetime):
        # ctime objects use time.strftime which does not support the %f format
        # for microseconds, so convert to regular datetime object
        _time_obj = time_obj._to_real_datetime()
    result = str(_time_obj.strftime(specs.TIMEFMT))
    return result


def write_variable(var_id: str,
                   granule: Granule,
                   idf_file: netCDF4.Dataset,
                   specs: typing.Any) -> WriteVariableResult:
    """"""
    var_dict = granule.vars[var_id]

    logger.debug(f'write variable {var_id}')

    # Options
    zlib = var_dict['options'].get('zlib', True)
    complevel = var_dict['options'].get('complevel', 4)
    shuffle = var_dict['options'].get('shuffle', True)
    fletcher32 = var_dict['options'].get('fletcher32', False)
    contiguous = var_dict['options'].get('contiguous', False)
    chunksizes = var_dict['options'].get('chunksizes', None)
    endian = var_dict['options'].get('endian', 'native')
    ls_digit = var_dict['options'].get('least_significant_digit', None)

    # Create variable
    dims = var_dict['options'].get('dimensions', None)
    if dims is None:
        dims = granule.get_data_model_dimensions()
        extra_dims = var_dict['options'].get('extra_dimensions', None)
        if extra_dims is not None:
            dims = dims + list(extra_dims)
    var_name = var_dict['name']
    dtype = var_dict['array'].dtype
    fill_value = dtype.type(var_dict['_FillValue'])
    _value = idf_file.createVariable(var_name, dtype, dims,
                                     zlib=zlib, complevel=complevel,
                                     shuffle=shuffle, fletcher32=fletcher32,
                                     contiguous=contiguous,
                                     chunksizes=chunksizes, endian=endian,
                                     least_significant_digit=ls_digit,
                                     fill_value=fill_value)

    # Assign data
    value = var_dict['array']
    if len(dims) == (1 + len(value.shape)):
        _value[:] = value[numpy.newaxis, :]
    else:
        _value[:] = value[:]

    if 'add_offset' in var_dict.keys():
        _value.add_offset = numpy.float32(var_dict['add_offset'])
    if 'scale_factor' in var_dict.keys():
        _value.scale_factor = numpy.float32(var_dict['scale_factor'])

    processed_var_attrs = ['name', 'add_offset', 'scale_factor', '_FillValue',
                           'options', 'dimensions', 'array', 'datatype']

    for attr in specs.VARIABLE_MANDATORY_ATTRIBUTES:
        if attr in var_dict.keys() and not hasattr(_value, attr):
            value = var_dict[attr]
            if attr in specs.VARIABLE_TEXT_ATTRIBUTES:
                setattr(_value, attr, str(value).encode('utf-8'))
            else:
                setattr(_value, attr, value)
            processed_var_attrs.append(attr)

    # CF attributes
    for attr in specs.VARIABLE_CF_ATTRIBUTES:
        if attr in var_dict.keys():
            value = var_dict[attr]

            # Make sure valid_min and valid_max use the right type
            if attr in ('valid_min', 'valid_max'):
                value = dtype.type(value)
                setattr(_value, attr, value)
            elif attr in specs.VARIABLE_TEXT_ATTRIBUTES:
                setattr(_value, attr, str(value).encode('utf-8'))
            else:
                setattr(_value, attr, value)
            processed_var_attrs.append(attr)

    # Optional attributes
    for attr in specs.VARIABLE_OPTIONAL_ATTRIBUTES:
        if attr in var_dict:
            value = var_dict[attr]
            if attr in specs.VARIABLE_TEXT_ATTRIBUTES:
                setattr(_value, attr, str(value).encode('utf-8'))
            else:
                setattr(_value, attr, value)
            processed_var_attrs.append(attr)

    return (_value, processed_var_attrs)


def write_global_attributes(input_opts: InputOptions,
                            output_opts: OutputOptions,
                            granule: Granule,
                            idf_file: netCDF4.Dataset,
                            specs: typing.Any) -> typing.List[str]:
    """Write global attributes to an IDF file. Attributes not defined in the
    IDF format specifications will be ignored.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.types.Granule
        Object representing the granule data and metadata
    idf_file: netCDF4.Dataset
        Handle of the IDF file to edit
    specs: module
        Module which contains the specifications for the IDF format

    Returns
    -------
    list
        The list of attributes that the method has included in the IDF file
    """
    # Mandatory attributes
    start_dt = _to_spec_time_format(granule.meta['time_coverage_start'], specs)
    stop_dt = _to_spec_time_format(granule.meta['time_coverage_end'], specs)

    spatial_resolution = numpy.float32(granule.meta['idf_spatial_resolution'])
    subsampling_factor = numpy.int32(0)  # Hardcoded
    spatial_resolution_units = b'm'  # Hardcoded

    logger.warning('idf_subsampling_factor value is hardcoded to "0"')
    logger.warning('idf_spatial_resolution_units value is hardcoded to "m"')

    idf_file.idf_version = granule.idf_version.encode('utf-8')
    idf_file.idf_granule_id = granule.meta['idf_granule_id'].encode('utf-8')
    idf_file.time_coverage_start = start_dt.encode('utf-8')
    idf_file.time_coverage_end = stop_dt.encode('utf-8')
    idf_file.idf_subsampling_factor = subsampling_factor
    idf_file.idf_spatial_resolution = spatial_resolution
    idf_file.idf_spatial_resolution_units = spatial_resolution_units

    processed_attrs = ['idf_version', 'idf_granule_id', 'time_coverage_start',
                       'time_coverage_end', 'idf_subsampling_factor',
                       'idf_spatial_resolution',
                       'idf_spatial_resolution_units']

    for attr in specs.MANDATORY_ATTRIBUTES:
        if attr in granule.meta and attr not in processed_attrs:
            if attr in specs.NUMERICAL_ATTRIBUTES:
                setattr(idf_file, attr, _to_numerical(granule.meta[attr]))
            elif attr in specs.TEXT_ATTRIBUTES:
                setattr(idf_file, attr,
                        str(granule.meta[attr]).encode('utf-8'))
            else:
                setattr(idf_file, attr, granule.meta[attr])
            processed_attrs.append(attr)

    # Supported attributes
    for attr in specs.SUPPORTED_ATTRIBUTES:
        if attr in granule.meta:
            if attr in specs.NUMERICAL_ATTRIBUTES:
                setattr(idf_file, attr, _to_numerical(granule.meta[attr]))
            elif attr in specs.TEXT_ATTRIBUTES:
                setattr(idf_file, attr,
                        str(granule.meta[attr]).encode('utf-8'))
            else:
                setattr(idf_file, attr, granule.meta[attr])
            processed_attrs.append(attr)

    # CF attributes
    idf_software = f'{idf_converter.__package__} {idf_converter_version}'
    _processing_software = granule.meta.get('processing_software', None)
    if _processing_software is None or 0 <= len(_processing_software):
        processing_software = idf_software
    else:
        processing_software = f'{_processing_software}\n{idf_software}'
    idf_file.processing_software = processing_software.encode('utf-8')
    processed_attrs.append('processing_software')

    now = datetime.datetime.utcnow()
    idf_history = f'{now:%Y-%m-%dT%H:%M:%SZ} Conversion to IDF'
    _history = granule.meta.get('history', None)
    if _history is None or 0 <= len(_history):
        history = idf_history
    else:
        history = f'{_history}\n{idf_history}'
    idf_file.history = history.encode('utf-8')
    processed_attrs.append('history')

    for attr in specs.CF_ATTRIBUTES:
        if attr in granule.meta and attr not in processed_attrs:
            if attr in specs.NUMERICAL_ATTRIBUTES:
                setattr(idf_file, attr, _to_numerical(granule.meta[attr]))
            elif attr in specs.TEXT_ATTRIBUTES:
                setattr(idf_file, attr,
                        str(granule.meta[attr]).encode('utf-8'))
            else:
                setattr(idf_file, attr, granule.meta[attr])
            processed_attrs.append(attr)

    # Optional attributes
    idf_file.netcdf_version_id = netCDF4.getlibversion().encode('utf-8')
    processed_attrs.append('netcdf_version_id')

    for attr in specs.OPTIONAL_ATTRIBUTES:
        if attr in granule.meta and attr not in processed_attrs:
            if attr in specs.NUMERICAL_ATTRIBUTES:
                setattr(idf_file, attr, _to_numerical(granule.meta[attr]))
            elif attr in specs.TEXT_ATTRIBUTES:
                setattr(idf_file, attr,
                        str(granule.meta[attr]).encode('utf-8'))
            else:
                setattr(idf_file, attr, granule.meta[attr])
            processed_attrs.append(attr)

    if not hasattr(idf_file, 'license'):
        _license = b'IDF protocol describes data use as free and open'
        idf_file.license = _license
        processed_attrs.append('license')

    if not hasattr(idf_file, 'standard_name_vocabulary'):
        _std_voc = b'NetCDF Climate and Forecast (CF) Metadata Convention'
        idf_file.standard_name_vocabulary = _std_voc
        processed_attrs.append('standard_name_vocabulary')

    # Special attributes
    for attr in specs.SPECIAL_ATTRIBUTES:
        if attr in granule.meta and attr not in processed_attrs:
            if attr in specs.NUMERICAL_ATTRIBUTES:
                setattr(idf_file, attr, _to_numerical(granule.meta[attr]))
            elif attr in specs.TEXT_ATTRIBUTES:
                setattr(idf_file, attr,
                        str(granule.meta[attr]).encode('utf-8'))
            else:
                setattr(idf_file, attr, granule.meta[attr])
            processed_attrs.append(attr)

    return processed_attrs


def get_data_model(data_model: str) -> typing.Any:
    """"""
    mod = None
    _data_model = data_model.lower()
    if 'time_dependent' == _data_model:
        import idf_converter.lib.idf.v1.time_dependent
        mod = idf_converter.lib.idf.v1.time_dependent
    elif 'swath' == _data_model:
        import idf_converter.lib.idf.v1.swath
        mod = idf_converter.lib.idf.v1.swath
    elif 'grid_yx' == _data_model:
        import idf_converter.lib.idf.v1.grid_yx
        mod = idf_converter.lib.idf.v1.grid_yx
    elif 'grid_latlon' == _data_model:
        import idf_converter.lib.idf.v1.grid_latlon
        mod = idf_converter.lib.idf.v1.grid_latlon
    else:
        raise DataModelNotSupported(data_model)

    return mod


def granule2idf(input_opts: InputOptions,
                output_opts: OutputOptions,
                granule: Granule,
                idf_file: netCDF4.Dataset) -> Granule2IDFResult:
    """
    Create dimensions, variables and attributes required by the IDF format
    based on granule information.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.types.Granule
        Object representing the granule data and metadata
    idf_file: netCDF4.Dataset
        Handle of the IDF file to edit

    Returns
    -------
    tuple
        Information about granule information that the method has processed.
        The tuple contains three lists::
        - a list of names for the dimensions created by the method
        - a list of identifiers for the variables created by the method
        - a list of names for the attributes created by the method
    """
    data_model_mod = get_data_model(granule.data_model)
    format_granule = data_model_mod.format_granule

    dims_done, vars_done, attrs_done = format_granule(input_opts, output_opts,
                                                      granule, idf_file)

    specs = idf_converter.lib.get_idf_specs(granule.idf_version)
    processed_attrs = write_global_attributes(input_opts, output_opts, granule,
                                              idf_file, specs)
    attrs_done.extend(processed_attrs)

    return (dims_done, vars_done, attrs_done)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-10
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import numpy
import typing
import logging
import idf_converter.lib
import idf_converter.readers.netcdf_grid_latlon
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger()

DATA_MODEL = idf_converter.readers.netcdf_grid_latlon.DATA_MODEL


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    in_msg, out_msg = idf_converter.readers.netcdf_grid_latlon.help()
    return (in_msg, out_msg)


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """

    # Define default input options
    if 'time_variable' not in input_opts:
        input_opts['time_variable'] = 'time'
    if 'lon_variable' not in input_opts:
        input_opts['lon_variable'] = 'longitude'
    if 'lat_variable' not in input_opts:
        input_opts['lat_variable'] = 'latitude'
    if 'variables' not in input_opts:
        input_opts['variables'] = 'VMDR,VHM0,VTM02,VTM10,VTPK,VPED'
        input_opts['variables'] += ',VMDR_WW,VHM0_WW,VTM01_WW'
        input_opts['variables'] += ',VMDR_SW1,VHM0_SW1,VTM01_SW1'
        input_opts['variables'] += ',VMDR_SW2,VHM0_SW2,VTM01_SW2'
        input_opts['variables'] += ',VSDX,VSDY'
    if 'time_coverage_relative_start' not in input_opts:
        input_opts['time_coverage_relative_start'] = '0'
    if 'time_coverage_relative_end' not in input_opts:
        input_opts['time_coverage_relative_end'] = '10800'
    if 'global_overrides' not in input_opts:
        input_opts['global_overrides'] = 'processing_level:L4'
        input_opts['global_overrides'] += ',product_version:1.0'
        input_opts['global_overrides'] += ',file_version:1.0'

    grid_latlon = idf_converter.readers.netcdf_grid_latlon
    for result in grid_latlon.read_data(input_opts, output_opts):
        input_opts, output_opts, granule, _transforms = result

        # The transforms list returned by the lat/lon grid reader is shared by
        # all the subgranules, i.e. modifying this list in a loop has
        # side-effects for the subsequent iterations.
        # Create a brand new transforms list to avoid this problem.
        transforms = []

        directions = [x for x in ('VMDR', 'VMDR_SW1', 'VMDR_SW2', 'VMDR_WW')
                      if x in granule.vars.keys()]
        if 0 < len(directions):
            dir_cfg = {x: {'direction': x,
                           'module': None,
                           'radians': False,
                           'angle_to_east': numpy.pi / 2.0,
                           'clockwise': True,
                           'meteo': True} for x in directions}
            transforms.append(('dir2vectors', {'targets': directions,
                                               'configs': dir_cfg}))

        # Add eastward_ and northward_ variables to the export targets
        targets = output_opts['__export']
        for dir_var_name in directions:
            targets.append(f'eastward_{dir_var_name}')
            targets.append(f'northward_{dir_var_name}')
        output_opts['__export'] = targets

        yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-07
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
import idf_converter.readers.netcdf_grid_latlon
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = idf_converter.readers.netcdf_grid_latlon.DATA_MODEL


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\t'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    # Read variables
    f_handler = netCDF4.Dataset(input_path, 'r')
    title = f_handler.title
    f_handler.close()

    _, _, _, _, _, year_str, month_str, _ = title.strip().split(' ')
    year = int(year_str)
    month = ['january', 'february', 'march',
             'april', 'may', 'june',
             'july', 'august', 'september',
             'october', 'november', 'december'].index(month_str.lower()) + 1
    indicative_dt = datetime.datetime(year, month, 15, 12, 0, 0)

    input_opts['lon_variable'] = 'LONGITUDE'
    input_opts['lat_variable'] = 'LATITUDE'
    input_opts['depth_variable'] = 'LEVEL'
    input_opts['depth_dimension'] = 'LEVEL'
    input_opts['variables'] = 'TEMP,SALT,PTEMP,PDEN,ADDEP,SPICE'
    input_opts['time_coverage_relative_start'] = 'month_beginning'
    input_opts['time_coverage_relative_end'] = 'month_end'
    input_opts['indicative_time'] = f'{indicative_dt:%Y-%m-%dT%H:%M:%SZ}'

    generic = idf_converter.readers.netcdf_grid_latlon.read_data(input_opts,
                                                                 output_opts)
    input_opts, output_opts, granule, transforms = next(generic)

    yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-07
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import struct
import typing
import logging
import datetime
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = 'GRID_LATLON'


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\t'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    # Read variables
    with open(input_path, 'rb') as f:
        file_content = f.read()

    fmt = '<{}f'.format(1440*480*8)
    _all_data = struct.unpack(fmt, file_content)
    all_data = numpy.reshape(_all_data, (8, 480, 1440,))
    all_data = numpy.roll(all_data, 720, axis=2)

    granule_fname = os.path.basename(input_path)
    _granule_name = granule_fname  # decompressed binary data have no extension
    _, granule_date = _granule_name.rsplit('_', 1)  # %Y%m%d
    _, version, process, _ = _granule_name.split('_', 3)

    input_opts['geoloc_at_pixel_center'] = 'no'
    dim1 = numpy.array([-60.0 + i * 0.25 for i in range(0, 480)])
    dim2 = numpy.array([-180.0 + i * 0.25 for i in range(0, 1440)])

    granule.vars['lat'] = {'array': dim1,
                           'units': 'degrees_north',
                           'datatype': dim1.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': dim2,
                           'units': 'degrees_east',
                           'datatype': dim2.dtype,
                           'options': {}}

    granule.dims['lat'] = dim1.size
    granule.dims['lon'] = dim2.size

    granule.vars['rain'] = {'name': 'rain',
                            'valid_min': 0.0,
                            'valid_max': 75,
                            'options': {}}

    granule.meta['idf_spatial_resolution'] = 27500  # ~ 1/4 degree
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['institution'] = 'NOAA CPC'

    dtime_base = datetime.datetime.strptime(granule_date, '%Y%m%d')
    for i in range(0, 8):
        hour = 3 * i
        granule_name = '{}_{:02d}'.format(_granule_name, hour)

        start_dt = dtime_base + datetime.timedelta(hours=hour)
        stop_dt = start_dt + datetime.timedelta(hours=3)

        granule.meta['idf_granule_id'] = granule_name
        granule.meta['time_coverage_start'] = start_dt
        granule.meta['time_coverage_end'] = stop_dt

        # input contains 3-hourly rate, we want the amount of precipitation
        # during these 3 hours
        _data = all_data[i, :, :] * 3
        granule.vars['rain']['array'] = _data

        transforms = []

        # Mask where there are no precipitations
        mask = (_data <= 0.0)  # nodata is -999
        transforms.append(('static_common_mask', {'targets': ('rain',),
                                                  'mask': mask}))

        output_opts['__export'] = ['rain', ]

        yield input_opts, output_opts, granule, transforms



# vim: ts=4:sts=4:sw=4
#
# @date 2024-08-23
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2024 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import typing
import logging

import numpy
import numpy.typing
from scipy.ndimage import gaussian_filter

from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)


def compute_swot_roughness(input_opts: InputOptions,
                           output_opts: OutputOptions,
                           granule: Granule,
                           *args: typing.Iterable[typing.Any],
                           targets: typing.Sequence[str],
                           ) -> FormatterResult:
    """"""
    # Inputs
    sigma0 = granule.vars['sigma0']['array']
    sigma0_fill = granule.vars['sigma0']['_FillValue']
    sigma0_data = numpy.ma.getdata(sigma0)
    sigma0_mask = numpy.ma.getmaskarray(sigma0)

    sigp = sigma0_data + 2
    mask = (sigma0_mask | (sigma0_data < 10) | (sigma0_data > 40))
    sigp[numpy.where(mask)] = numpy.nan

    # Avoid runtime warning regarding mean computation on empty slices for
    # columns that contain only NaN values
    profil_mask = (numpy.isnan(sigp).all(axis=0))
    sigp[0, numpy.where(profil_mask)] = 1

    profil = numpy.nanmean(numpy.log(sigp),
                           axis=0)

    # Columns containing only NaN values should have a NoN profil
    profil[profil_mask] = numpy.nan

    # Log can be computed without warning only if sigma0 + 2 > 0
    # When sigma0 == 2, log = -inf
    # When sigma0 < -2, log = nan
    log_sigma0_2_ok_mask = ((~sigma0_mask)
                            & (sigma0_data > -2))
    log_sigma0_2 = numpy.log(sigma0_data + 2,
                             out=numpy.ones_like(sigma0_data,
                                                 dtype=numpy.float64),
                             where=log_sigma0_2_ok_mask)

    # Avoid divide by 0 errors where log_sigma0_2 is null, i.e. where sigma0
    # is -1
    log_sigma0_2_null_mask = (log_sigma0_2 == 0)
    log_sigma0_2[numpy.where(log_sigma0_2_null_mask)] = 1

    ssr = profil[numpy.newaxis, :] / log_sigma0_2

    # ssr should be infinite for pixels where log_sigma0_2 was 0
    ssr[numpy.where(log_sigma0_2_null_mask)] = numpy.inf

    # log_sigma0_2 should be -inf where sigma0 == -2, so ssr should be 0 for
    # these pixels
    ssr[numpy.where(sigma0_data == -2)] = 0

    # Mask where log_sigma0_2 is undefined, i.e. where sigma0 was masked and
    # where sigma0 < -2, as it leads to log(<0)
    ssr = numpy.ma.masked_where((sigma0_mask | (sigma0_data < -2)), ssr)

    granule.vars['roughness']['array'] = ssr
    granule.vars['roughness']['_FillValue'] = sigma0_fill

    return input_opts, output_opts, granule


def compute_swot_ssha_unedited_nolr(input_opts: InputOptions,
                                    output_opts: OutputOptions,
                                    granule: Granule,
                                    *args: typing.Iterable[typing.Any],
                                    targets: typing.Sequence[str],
                                    ) -> FormatterResult:
    """"""
    ssha_unedited = granule.vars['ssha_unedited']['array']
    ssha_unedited_fill = granule.vars['ssha_unedited']['_FillValue']
    ssha_unedited_data = numpy.ma.getdata(ssha_unedited)
    ssha_unedited_mask = numpy.ma.getmaskarray(ssha_unedited)

    vmask = (ssha_unedited_mask
             | (ssha_unedited_data == ssha_unedited_fill)
             | (numpy.abs(ssha_unedited_data) > 100))

    mask = (~vmask).astype(numpy.float32)
    ssha_unedited_data[numpy.where(vmask)] = 0

    gf = gaussian_filter(ssha_unedited_data * mask, sigma=10)

    gaussian_mask = gaussian_filter(mask, sigma=10)

    # Prevent divide by 0
    gaussian_mask_0 = (gaussian_mask == 0)
    gaussian_mask[numpy.where(gaussian_mask_0)] = 1
    gf /= gaussian_mask
    gf[numpy.where(gaussian_mask_0)] = 0

    nolr = ssha_unedited_data - gf
    nolr_mask = (vmask | (nolr > 40))
    nolr = numpy.ma.masked_where(nolr_mask, nolr)

    granule.vars['ssha_unedited_nolr']['array'] = nolr
    granule.vars['ssha_unedited_nolr']['_FillValue'] = ssha_unedited_fill

    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-09
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import numpy
import typing
import logging
from idf_converter.lib import mod2pi
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)

DirConfig = typing.Dict[str, typing.Dict[str, typing.Any]]


class MissingVectorConfig(ValueError):
    """Error raised when it is not possible to compute one of the requested
    vectors because its configuration has not been provided."""
    def __init__(self, name: str) -> None:
        self.name = name


class MissingDirectionVariable(ValueError):
    """Error raised when the vector configuration does not define the name of
    the variable which contains direction data."""
    pass


class MissingDirectionData(Exception):
    """Error raised when the variable for the vector direction is not
    available in the granule."""
    def __init__(self, var_name: str) -> None:
        """"""
        self.var_name = var_name


class MissingModuleData(Exception):
    """Error raised when the variable for the vector module is not available
    in the granule."""
    def __init__(self, var_name: str) -> None:
        """"""
        self.var_name = var_name


def from_dir(input_opts: InputOptions,
             output_opts: OutputOptions,
             granule: Granule,
             *args: typing.Iterable[typing.Any],
             targets: typing.Iterable[str],
             configs: DirConfig) -> FormatterResult:
    """"""
    for target in targets:
        cfg = configs.get(target, None)
        if cfg is None:
            raise MissingVectorConfig(target)

        direction_var_name = cfg.get('direction', None)
        if direction_var_name is None:
            raise MissingDirectionVariable()

        has_extra_values = 'extra_values' in input_opts.keys()

        if direction_var_name in granule.vars.keys():
            dir_data = granule.vars[direction_var_name]['array']
        elif has_extra_values:
            if direction_var_name in input_opts['extra_values'].keys():
                dir_var = input_opts['extra_values'][direction_var_name]
                dir_data = dir_var['array']
            else:
                raise MissingDirectionData(direction_var_name)
        else:
            raise MissingDirectionData(direction_var_name)

        units = None
        module_data = None
        module_var_name = cfg.get('module', None)
        if module_var_name is not None:
            if module_var_name in granule.vars.keys():
                units = granule.vars[module_var_name].get('units', '')
                module_data = granule.vars[module_var_name]['array']
            elif has_extra_values:
                if module_var_name in input_opts['extra_values'].keys():
                    module_var = input_opts['extra_values'][module_var_name]
                    units = module_var.get('units', '')
                    module_data = module_var['array']
                else:
                    raise MissingModuleData(module_var_name)
            else:
                raise MissingModuleData(module_var_name)
        else:
            units = ''
            module_data = numpy.ones(numpy.shape(dir_data))

        radians = cfg.get('radians', False)
        if not radians:
            dir_data = numpy.deg2rad(dir_data)

        is_meteo_convention = cfg.get('meteo', False)
        if is_meteo_convention:
            # Convert "direction from" to "direction to"
            dir_data, trunc_to_0 = mod2pi(numpy.pi + dir_data, 9)
            if trunc_to_0 is True:
                logger.warning('Some vector angle values very close to 0 have '
                               'been truncated to 0 to avoid underflow issues')

        # Convert to counter-clockwise from east
        angle_to_east = cfg.get('angle_to_east', 0.0)
        rotation_sign = -1.0 if cfg.get('clockwise', False) else 1.0
        _dir_data = angle_to_east + rotation_sign * dir_data
        try:
            dir_data = numpy.mod(_dir_data, 2.0 * numpy.pi,
                                 where=numpy.isfinite(_dir_data))
        except FloatingPointError:
            UFLOW_FIX = 10**6
            dir_data = numpy.mod(UFLOW_FIX * _dir_data,
                                 UFLOW_FIX * 2.0 * numpy.pi,
                                 where=numpy.isfinite(_dir_data))
            eps = numpy.finfo(dir_data.dtype).eps
            underflow_mask = ((dir_data > 0.0) & (dir_data < UFLOW_FIX * eps))
            if underflow_mask.any():
                logger.warning(f'Some direction angles lower than {eps}'
                               ' radians have been truncated to 0 to avoid'
                               ' underflow errors')
                dir_data[numpy.where(underflow_mask)] = 0.0
            dir_data = dir_data / UFLOW_FIX

        try:
            eastward_data = module_data * numpy.cos(dir_data)
        except FloatingPointError:
            module_data_sq = module_data * module_data
            dir_data_sin = numpy.sin(dir_data)
            module_data_sin_sq = module_data_sq * dir_data_sin * dir_data_sin
            sign = numpy.ones(dir_data.shape)
            negative_mask_min = (dir_data > numpy.pi / 2.0)
            negative_mask_max = (dir_data < 3.0 * numpy.pi / 2.0)
            sign[numpy.where(negative_mask_min & negative_mask_max)] = -1.0
            eastward_module = numpy.sqrt(module_data_sq - module_data_sin_sq)
            eastward_data = sign * eastward_module

        try:
            northward_data = module_data * numpy.sin(dir_data)
        except FloatingPointError:
            module_data_sq = module_data * module_data
            dir_data_cos = numpy.cos(dir_data)
            module_data_cos_sq = module_data_sq * dir_data_cos * dir_data_cos
            sign = numpy.ones(dir_data.shape)
            sign[numpy.where(numpy.pi < dir_data)] = -1.0
            northward_module = numpy.sqrt(module_data_sq - module_data_cos_sq)
            northward_data = sign * northward_module

        vmax = numpy.abs(numpy.nanmax(module_data))
        vmin = -1.0 * vmax

        eastward_var_name = f'eastward_{target}'
        granule.vars[eastward_var_name] = {'array': eastward_data,
                                           'name': eastward_var_name,
                                           'units': units,
                                           'valid_min': vmin,
                                           'valid_max': vmax,
                                           'datatype': eastward_data.dtype,
                                           'options': {}}
        northward_var_name = f'northward_{target}'
        granule.vars[northward_var_name] = {'array': northward_data,
                                            'name': northward_var_name,
                                            'units': units,
                                            'valid_min': vmin,
                                            'valid_max': vmax,
                                            'datatype': northward_data.dtype,
                                            'options': {}}

    return input_opts, output_opts, granule
